@@ -38,13 +38,13 @@
 %%--------------------------------------------------------------------------
 
 -record(v1, {parent, sock, connection, callback, recv_len, pending_recv,
-             connection_state, queue_collector, heartbeater, ch_sup3_pid,
-             channel_sup_sup_pid, start_heartbeat_fun, buf, buf_len, throttle}).
+             connection_state, queue_collector, heartbeater, helper_sup,
+             channel_sup_sup_pid, buf, buf_len, throttle}).
 
 -record(connection, {user, timeout_sec, frame_max, auth_mechanism, auth_state,
                      hostname}).
 
--record(throttle, {conserve_resources, last_blocked_by, last_blocked_at}).
+-record(throttle, {alarmed_by, last_blocked_by, last_blocked_at}).
 
 -define(IS_RUNNING(State),
         (State#v1.connection_state =:= running orelse
@@ -53,23 +53,22 @@
 
 %%--------------------------------------------------------------------------
 
-unpack_from_0_9_1({Parent, Sock,RecvLen, PendingRecv, QueueCollector,
-                   ChSup3Pid, SHF, Buf, BufLen}) ->
+unpack_from_0_9_1({Parent, Sock,RecvLen, PendingRecv,
+                   HelperSupPid, Buf, BufLen}) ->
     #v1{parent              = Parent,
         sock                = Sock,
         callback            = handshake,
         recv_len            = RecvLen,
         pending_recv        = PendingRecv,
         connection_state    = pre_init,
-        queue_collector     = QueueCollector,
+        queue_collector     = undefined,
         heartbeater         = none,
-        ch_sup3_pid         = ChSup3Pid,
-        start_heartbeat_fun = SHF,
+        helper_sup          = HelperSupPid,
         buf                 = Buf,
         buf_len             = BufLen,
-        throttle = #throttle{conserve_resources = false,
-                             last_blocked_by    = none,
-                             last_blocked_at    = never},
+        throttle = #throttle{alarmed_by      = [],
+                             last_blocked_by = none,
+                             last_blocked_at = never},
         connection = #connection{user           = none,
                                  timeout_sec    = ?HANDSHAKE_TIMEOUT,
                                  frame_max      = ?FRAME_MIN_SIZE,
@@ -88,8 +87,8 @@ system_terminate(Reason, _Parent, _Deb, _State) ->
 system_code_change(Misc, _Module, _OldVsn, _Extra) ->
     {ok, Misc}.
 
-conserve_resources(Pid, _Source, Conserve) ->
-    Pid ! {conserve_resources, Conserve},
+conserve_resources(Pid, Source, Conserve) ->
+    Pid ! {conserve_resources, Source, Conserve},
     ok.
 
 server_properties() ->
@@ -151,9 +150,14 @@ mainloop(Deb, State = #v1{sock = Sock, buf = Buf, buf_len = BufLen}) ->
             end
     end.
 
-handle_other({conserve_resources, Conserve},
-             State = #v1{throttle = Throttle}) ->
-    Throttle1 = Throttle#throttle{conserve_resources = Conserve},
+handle_other({conserve_resources, Source, Conserve},
+             State = #v1{throttle = Throttle =
+                             #throttle{alarmed_by = CR}}) ->
+    CR1 = case Conserve of
+              true  -> lists:usort([Source | CR]);
+              false -> CR -- [Source]
+          end,
+    Throttle1 = Throttle#throttle{alarmed_by = CR1},
     control_throttle(State#v1{throttle = Throttle1});
 handle_other({'EXIT', Parent, Reason}, State = #v1{parent = Parent}) ->
     terminate(io_lib:format("broker forced connection closure "
@@ -210,8 +214,9 @@ terminate(_Reason, State) ->
     {force, State}.
 
 control_throttle(State = #v1{connection_state = CS, throttle = Throttle}) ->
-    case {CS, (Throttle#throttle.conserve_resources orelse
-               credit_flow:blocked())} of
+    IsThrottled = ((Throttle#throttle.alarmed_by =/= []) orelse
+               credit_flow:blocked()),
+    case {CS, IsThrottled} of
         {running,   true} -> State#v1{connection_state = blocking};
         {blocking, false} -> State#v1{connection_state = running};
         {blocked,  false} -> ok = rabbit_heartbeat:resume_monitor(
@@ -222,10 +227,10 @@ control_throttle(State = #v1{connection_state = CS, throttle = Throttle}) ->
         {_,            _} -> State
     end.
 
-update_last_blocked_by(Throttle = #throttle{conserve_resources = true}) ->
-    Throttle#throttle{last_blocked_by = resource};
-update_last_blocked_by(Throttle = #throttle{conserve_resources = false}) ->
-    Throttle#throttle{last_blocked_by = flow}.
+update_last_blocked_by(Throttle = #throttle{alarmed_by = []}) ->
+    Throttle#throttle{last_blocked_by = flow};
+update_last_blocked_by(Throttle) ->
+    Throttle#throttle{last_blocked_by = resource}.
 
 %%--------------------------------------------------------------------------
 %% error handling / termination
@@ -354,10 +359,10 @@ handle_1_0_connection_frame(#'v1_0.open'{ max_frame_size = ClientFrameMax,
                                           hostname = Hostname,
                                           properties = Props },
                             State = #v1{
-                              start_heartbeat_fun = SHF,
                               connection_state = starting,
                               connection = Connection,
                               throttle   = Throttle,
+                              helper_sup = HelperSupPid,
                               sock = Sock}) ->
     ClientProps        = case Props of
                              undefined -> [];
@@ -382,6 +387,8 @@ handle_1_0_connection_frame(#'v1_0.open'{ max_frame_size = ClientFrameMax,
                                "frame_max=~w < ~w min size",
                                [FrameMax, ?FRAME_1_0_MIN_SIZE]);
            true ->
+                {ok, Collector} =
+                    rabbit_connection_helper_sup:start_queue_collector(HelperSupPid),
                 SendFun =
                     fun() ->
                             Frame =
@@ -398,13 +405,16 @@ handle_1_0_connection_frame(#'v1_0.open'{ max_frame_size = ClientFrameMax,
                 %%         actual timeout threshold
                 ReceiverHeartbeatSec = lists:min([HeartbeatSec * 2, 4294967]),
                 %% TODO: only start heartbeat receive timer at next next frame
-                Heartbeater = SHF(Sock, ClientHeartbeatSec, SendFun,
-                                  ReceiverHeartbeatSec, ReceiveFun),
+                Heartbeater =
+                    rabbit_heartbeat:start(HelperSupPid, Sock,
+                                           ClientHeartbeatSec, SendFun,
+                                           ReceiverHeartbeatSec, ReceiveFun),
                 State#v1{connection_state = running,
                          connection = Connection#connection{
                                                    frame_max = FrameMax,
                                                    hostname  = Hostname},
-                         heartbeater = Heartbeater}
+                         heartbeater = Heartbeater,
+                         queue_collector = Collector}
         end,
     %% TODO enforce channel_max
     ok = send_on_channel0(
@@ -416,8 +426,7 @@ handle_1_0_connection_frame(#'v1_0.open'{ max_frame_size = ClientFrameMax,
                         properties     = server_properties()}),
     Conserve = rabbit_alarm:register(self(), {?MODULE, conserve_resources, []}),
     control_throttle(
-      State1#v1{throttle = Throttle#throttle{
-                             conserve_resources = Conserve}});
+      State1#v1{throttle = Throttle#throttle{alarmed_by = Conserve}});
 
 handle_1_0_connection_frame(_Frame, State) ->
     maybe_close(State#v1{connection_state = closing}).
@@ -574,14 +583,14 @@ start_1_0_connection(amqp,
             start_1_0_connection0(amqp, State)
     end.
 
-start_1_0_connection0(Mode, State = #v1{connection   = Connection,
-                                        ch_sup3_pid  = ChSup3Pid}) ->
+start_1_0_connection0(Mode, State = #v1{connection = Connection,
+                                        helper_sup = HelperSup}) ->
     ChannelSupSupPid =
         case Mode of
             sasl -> undefined;
             amqp -> {ok, Pid} =
                         supervisor2:start_child(
-                          ChSup3Pid,
+                          HelperSup,
                           {channel_sup_sup,
                            {rabbit_amqp1_0_session_sup_sup, start_link, []},
                            intrinsic, infinity, supervisor,
@@ -680,7 +689,7 @@ send_to_new_1_0_session(Channel, Frame, State) ->
 vhost({utf8, <<"vhost:", VHost/binary>>}) ->
     VHost;
 vhost(_) ->
-    {ok, DefaultVHost} = application:get_env(rabbit, default_vhost),
+    {ok, DefaultVHost} = application:get_env(default_vhost),
     DefaultVHost.
 
 %% End 1-0
