@@ -26,31 +26,12 @@
 -export([description/0]).
 -export([check_user_login/2, check_vhost_access/2, check_resource_access/3]).
 
--behaviour(gen_server).
-
--export([start_link/0]).
--export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2,
-         code_change/3]).
-
--define(SERVER, ?MODULE).
--define(L(F, A, S),  log("LDAP "         ++ F, A, S)).
--define(L1(F, A, S), log("    LDAP "     ++ F, A, S)).
--define(L2(F, A, S), log("        LDAP " ++ F, A, S)).
+-define(L(F, A),  log("LDAP "         ++ F, A)).
+-define(L1(F, A), log("    LDAP "     ++ F, A)).
+-define(L2(F, A), log("        LDAP " ++ F, A)).
 
 -import(rabbit_misc, [pget/2]).
 
--record(state, { servers,
-                 user_dn_pattern,
-                 dn_lookup_attribute,
-                 dn_lookup_base,
-                 other_bind,
-                 vhost_access_query,
-                 resource_access_query,
-                 tag_queries,
-                 use_ssl,
-                 log,
-                 port }).
-
 -record(impl, { user_dn, password }).
 
 %%--------------------------------------------------------------------
@@ -59,105 +40,152 @@ description() ->
     [{name, <<"LDAP">>},
      {description, <<"LDAP authentication / authorisation">>}].
 
-start_link() ->
-    gen_server:start_link({local, ?MODULE}, ?MODULE, [], []).
-
 %%--------------------------------------------------------------------
 
 check_user_login(Username, []) ->
-    gen_server:call(?SERVER, {login, Username}, infinity);
+    %% Without password, e.g. EXTERNAL
+    ?L("CHECK: passwordless login for ~s", [Username]),
+    R = with_ldap(creds(none),
+                  fun(LDAP) -> do_login(Username, none, LDAP) end),
+    ?L("DECISION: passwordless login for ~s: ~p",
+       [Username, log_result(R)]),
+    R;
 
 check_user_login(Username, [{password, Password}]) ->
-    gen_server:call(?SERVER, {login, Username, Password}, infinity);
+    ?L("CHECK: login for ~s", [Username]),
+    R = with_ldap({fill_user_dn_pattern(Username), Password},
+                  fun(LDAP) -> do_login(Username, Password, LDAP) end),
+    ?L("DECISION: login for ~s: ~p", [Username, log_result(R)]),
+    R;
 
 check_user_login(Username, AuthProps) ->
     exit({unknown_auth_props, Username, AuthProps}).
 
 check_vhost_access(User = #user{username = Username,
                                 impl     = #impl{user_dn = UserDN}}, VHost) ->
-    gen_server:call(?SERVER, {check_vhost, [{username, Username},
-                                            {user_dn,  UserDN},
-                                            {vhost,    VHost}], User},
-                    infinity).
+    Args = [{username, Username},
+            {user_dn,  UserDN},
+            {vhost,    VHost}],
+    ?L("CHECK: ~s for ~s", [log_vhost(Args), log_user(User)]),
+    R = evaluate_ldap(env(vhost_access_query), Args, User),
+    ?L("DECISION: ~s for ~s: ~p",
+       [log_vhost(Args), log_user(User), log_result(R)]),
+    R.
 
 check_resource_access(User = #user{username = Username,
                                    impl     = #impl{user_dn = UserDN}},
                       #resource{virtual_host = VHost, kind = Type, name = Name},
                       Permission) ->
-    gen_server:call(?SERVER, {check_resource, [{username,   Username},
-                                               {user_dn,    UserDN},
-                                               {vhost,      VHost},
-                                               {resource,   Type},
-                                               {name,       Name},
-                                               {permission, Permission}], User},
-                    infinity).
+    Args = [{username,   Username},
+            {user_dn,    UserDN},
+            {vhost,      VHost},
+            {resource,   Type},
+            {name,       Name},
+            {permission, Permission}],
+    ?L("CHECK: ~s for ~s", [log_resource(Args), log_user(User)]),
+    R = evaluate_ldap(env(resource_access_query), Args, User),
+    ?L("DECISION: ~s for ~s: ~p",
+       [log_resource(Args), log_user(User), log_result(R)]),
+    R.
 
 %%--------------------------------------------------------------------
 
-evaluate(Query, Args, User, LDAP, State) ->
-    ?L1("evaluating query: ~p", [Query], State),
-    evaluate0(Query, Args, User, LDAP, State).
+evaluate(Query, Args, User, LDAP) ->
+    ?L1("evaluating query: ~p", [Query]),
+    evaluate0(Query, Args, User, LDAP).
 
-evaluate0({constant, Bool}, _Args, _User, _LDAP, State) ->
-    ?L1("evaluated constant: ~p", [Bool], State),
+evaluate0({constant, Bool}, _Args, _User, _LDAP) ->
+    ?L1("evaluated constant: ~p", [Bool]),
     Bool;
 
-evaluate0({for, [{Type, Value, SubQuery}|Rest]}, Args, User, LDAP, State) ->
+evaluate0({for, [{Type, Value, SubQuery}|Rest]}, Args, User, LDAP) ->
     case pget(Type, Args) of
         undefined -> {error, {args_do_not_contain, Type, Args}};
-        Value     -> ?L1("selecting subquery ~s = ~s", [Type, Value], State),
-                     evaluate(SubQuery, Args, User, LDAP, State);
-        _         -> evaluate0({for, Rest}, Args, User, LDAP, State)
+        Value     -> ?L1("selecting subquery ~s = ~s", [Type, Value]),
+                     evaluate(SubQuery, Args, User, LDAP);
+        _         -> evaluate0({for, Rest}, Args, User, LDAP)
     end;
 
-evaluate0({for, []}, _Args, _User, _LDAP, _State) ->
+evaluate0({for, []}, _Args, _User, _LDAP) ->
     {error, {for_query_incomplete}};
 
-evaluate0({exists, DNPattern}, Args, _User, LDAP, State) ->
+evaluate0({exists, DNPattern}, Args, _User, LDAP) ->
     %% eldap forces us to have a filter. objectClass should always be there.
     Filter = eldap:present("objectClass"),
-    DN = fill(DNPattern, Args, State),
+    DN = fill(DNPattern, Args),
     R = object_exists(DN, Filter, LDAP),
-    ?L1("evaluated exists for \"~s\": ~p", [DN, R], State),
+    ?L1("evaluated exists for \"~s\": ~p", [DN, R]),
     R;
 
-evaluate0({in_group, DNPattern}, Args, User, LDAP, State) ->
-    evaluate({in_group, DNPattern, "member"}, Args, User, LDAP, State);
+evaluate0({in_group, DNPattern}, Args, User, LDAP) ->
+    evaluate({in_group, DNPattern, "member"}, Args, User, LDAP);
 
 evaluate0({in_group, DNPattern, Desc}, Args,
-          #user{impl = #impl{user_dn = UserDN}}, LDAP, State) ->
+          #user{impl = #impl{user_dn = UserDN}}, LDAP) ->
     Filter = eldap:equalityMatch(Desc, UserDN),
-    DN = fill(DNPattern, Args, State),
+    DN = fill(DNPattern, Args),
     R = object_exists(DN, Filter, LDAP),
-    ?L1("evaluated in_group for \"~s\": ~p", [DN, R], State),
+    ?L1("evaluated in_group for \"~s\": ~p", [DN, R]),
     R;
 
-evaluate0({match, StringQuery, REQuery}, Args, User, LDAP, State) ->
+evaluate0({'not', SubQuery}, Args, User, LDAP) ->
+    R = evaluate(SubQuery, Args, User, LDAP),
+    ?L1("negated result to ~s", [R]),
+    not R;
+
+evaluate0({'and', Queries}, Args, User, LDAP) when is_list(Queries) ->
+    R = lists:foldl(fun (Q,  true)  -> evaluate(Q, Args, User, LDAP);
+                        (_Q, false) -> false
+                    end, true, Queries),
+    ?L1("'and' result: ~s", [R]),
+    R;
+
+evaluate0({'or', Queries}, Args, User, LDAP) when is_list(Queries) ->
+    R = lists:foldl(fun (_Q, true)  -> true;
+                        (Q,  false) -> evaluate(Q, Args, User, LDAP)
+                    end, false, Queries),
+    ?L1("'or' result: ~s", [R]),
+    R;
+
+evaluate0({equals, StringQuery1, StringQuery2}, Args, User, LDAP) ->
+    safe_eval(fun (String1, String2) ->
+                      R = String1 =:= String2,
+                      ?L1("evaluated equals \"~s\", \"~s\": ~s",
+                          [String1, String2, R]),
+                      R
+              end,
+              evaluate(StringQuery1, Args, User, LDAP),
+              evaluate(StringQuery2, Args, User, LDAP));
+
+evaluate0({match, StringQuery, REQuery}, Args, User, LDAP) ->
     safe_eval(fun (String, RE) ->
                       R = case re:run(String, RE) of
                               {match, _} -> true;
                               nomatch    -> false
                           end,
                       ?L1("evaluated match \"~s\" against RE \"~s\": ~s",
-                          [String, RE, R], State),
+                          [String, RE, R]),
                       R
               end,
-              evaluate(StringQuery, Args, User, LDAP, State),
-              evaluate(REQuery, Args, User, LDAP, State));
+              evaluate(StringQuery, Args, User, LDAP),
+              evaluate(REQuery, Args, User, LDAP));
 
-evaluate0({string, StringPattern}, Args, _User, _LDAP, State) ->
-    R = fill(StringPattern, Args, State),
-    ?L1("evaluated string for \"~s\"", [R], State),
+evaluate0(StringPattern, Args, User, LDAP) when is_list(StringPattern) ->
+    evaluate0({string, StringPattern}, Args, User, LDAP);
+
+evaluate0({string, StringPattern}, Args, _User, _LDAP) ->
+    R = fill(StringPattern, Args),
+    ?L1("evaluated string for \"~s\"", [R]),
     R;
 
-evaluate0({attribute, DNPattern, AttributeName}, Args, _User, LDAP, State) ->
-    DN = fill(DNPattern, Args, State),
+evaluate0({attribute, DNPattern, AttributeName}, Args, _User, LDAP) ->
+    DN = fill(DNPattern, Args),
     R = attribute(DN, AttributeName, LDAP),
     ?L1("evaluated attribute \"~s\" for \"~s\": ~p",
-        [AttributeName, DN, R], State),
+        [AttributeName, DN, R]),
     R;
 
-evaluate0(Q, Args, _User, _LDAP, _State) ->
+evaluate0(Q, Args, _User, _LDAP) ->
     {error, {unrecognised_query, Q, Args}}.
 
 safe_eval(_F, {error, _}, _)          -> false;
@@ -192,20 +220,21 @@ attribute(DN, AttributeName, LDAP) ->
             E
     end.
 
-evaluate_ldap(Q, Args, User, State) ->
-    with_ldap(creds(User, State),
-              fun(LDAP) -> evaluate(Q, Args, User, LDAP, State) end, State).
+evaluate_ldap(Q, Args, User) ->
+    with_ldap(creds(User), fun(LDAP) -> evaluate(Q, Args, User, LDAP) end).
 
 %%--------------------------------------------------------------------
 
+with_ldap(Creds, Fun) -> with_ldap(Creds, Fun, env(servers)).
+
+with_ldap(_Creds, _Fun, undefined) ->
+    {error, no_ldap_servers_defined};
+
 %% TODO - ATM we create and destroy a new LDAP connection on every
 %% call. This could almost certainly be more efficient.
-with_ldap(Creds, Fun, State = #state{servers = Servers,
-                                     use_ssl = SSL,
-                                     log     = Log,
-                                     port    = Port}) ->
-    Opts0 = [{ssl, SSL}, {port, Port}],
-    Opts = case Log of
+with_ldap(Creds, Fun, Servers) ->
+    Opts0 = [{ssl, env(use_ssl)}, {port, env(port)}],
+    Opts = case env(log) of
                network ->
                    Pre = "    LDAP network traffic: ",
                    rabbit_log:info(
@@ -220,66 +249,66 @@ with_ldap(Creds, Fun, State = #state{servers = Servers,
         {ok, LDAP} ->
             try Creds of
                 anon ->
-                    ?L1("anonymous bind", [], State),
+                    ?L1("anonymous bind", []),
                     Fun(LDAP);
                 {UserDN, Password} ->
                     case eldap:simple_bind(LDAP, UserDN, Password) of
                         ok ->
-                            ?L1("bind succeeded: ~s", [UserDN], State),
+                            ?L1("bind succeeded: ~s", [UserDN]),
                             Fun(LDAP);
                         {error, invalidCredentials} ->
                             ?L1("bind returned \"invalid credentials\": ~s",
-                                [UserDN], State),
+                                [UserDN]),
                             {refused, UserDN, []};
                         {error, E} ->
-                            ?L1("bind error: ~s ~p", [UserDN, E], State),
+                            ?L1("bind error: ~s ~p", [UserDN, E]),
                             {error, E}
                     end
             after
                 eldap:close(LDAP)
             end;
         Error ->
-            ?L1("connect error: ~p", [Error], State),
+            ?L1("connect error: ~p", [Error]),
             Error
     end.
 
-get_env(F) ->
-    {ok, V} = application:get_env(F),
+env(F) ->
+    {ok, V} = application:get_env(rabbitmq_auth_backend_ldap, F),
     V.
 
-do_login(Username, Password, LDAP,
-         State = #state{ tag_queries = TagQueries }) ->
-    UserDN = username_to_dn(Username, LDAP, State),
+do_login(Username, Password, LDAP) ->
+    UserDN = username_to_dn(Username, LDAP),
     User = #user{username     = Username,
                  auth_backend = ?MODULE,
                  impl         = #impl{user_dn  = UserDN,
                                       password = Password}},
 
     TagRes = [begin
-                  ?L1("CHECK: does ~s have tag ~s?", [Username, Tag], State),
+                  ?L1("CHECK: does ~s have tag ~s?", [Username, Tag]),
                   R = evaluate(Q, [{username, Username},
-                                   {user_dn,  UserDN}], User, LDAP, State),
+                                   {user_dn,  UserDN}], User, LDAP),
                   ?L1("DECISION: does ~s have tag ~s? ~p",
-                      [Username, Tag, R], State),
+                      [Username, Tag, R]),
                   {Tag, R}
-              end || {Tag, Q} <- TagQueries],
+              end || {Tag, Q} <- env(tag_queries)],
     case [E || {_, E = {error, _}} <- TagRes] of
         []      -> {ok, User#user{tags = [Tag || {Tag, true} <- TagRes]}};
         [E | _] -> E
     end.
 
-username_to_dn(Username, _LDAP, State = #state{dn_lookup_attribute = none}) ->
-    fill_user_dn_pattern(Username, State);
+username_to_dn(Username, LDAP) ->
+    username_to_dn(Username, LDAP, env(dn_lookup_attribute)).
 
-username_to_dn(Username, LDAP, State = #state{dn_lookup_attribute = Attr,
-                                              dn_lookup_base      = Base}) ->
-    Filled = fill_user_dn_pattern(Username, State),
+username_to_dn(Username, _LDAP, none) ->
+    fill_user_dn_pattern(Username);
+
+username_to_dn(Username, LDAP, Attr) ->
+    Filled = fill_user_dn_pattern(Username),
     case eldap:search(LDAP,
-                      [{base, Base},
+                      [{base, env(dn_lookup_base)},
                        {filter, eldap:equalityMatch(Attr, Filled)},
                        {attributes, ["distinguishedName"]}]) of
-        {ok, #eldap_search_result{entries = [#eldap_entry{attributes = A}]}} ->
-            [DN] = pget("distinguishedName", A),
+        {ok, #eldap_search_result{entries = [#eldap_entry{object_name = DN}]}}->
             DN;
         {ok, #eldap_search_result{entries = Entries}} ->
             rabbit_log:warning("Searching for DN for ~s, got back ~p~n",
@@ -289,25 +318,27 @@ username_to_dn(Username, LDAP, State = #state{dn_lookup_attribute = Attr,
             exit(E)
     end.
 
-fill_user_dn_pattern(Username,
-                     State = #state{user_dn_pattern = UserDNPattern}) ->
-    fill(UserDNPattern, [{username, Username}], State).
+fill_user_dn_pattern(Username) ->
+    fill(env(user_dn_pattern), [{username, Username}]).
 
-creds(none, #state{other_bind = as_user}) ->
+creds(User) -> creds(User, env(other_bind)).
+
+creds(none, as_user) ->
     exit(as_user_no_password);
-creds(#user{impl = #impl{user_dn = UserDN, password = Password}},
-      #state{other_bind = as_user}) ->
+creds(#user{impl = #impl{user_dn = UserDN, password = Password}}, as_user) ->
     {UserDN, Password};
-creds(_, #state{other_bind = Creds}) ->
+creds(_, Creds) ->
     Creds.
 
-log(_Fmt, _Args, #state{log = false}) -> ok;
-log( Fmt,  Args, _State)              -> rabbit_log:info(Fmt ++ "~n", Args).
+log(Fmt,  Args) -> case env(log) of
+                       false -> ok;
+                       _     -> rabbit_log:info(Fmt ++ "~n", Args)
+                   end.
 
-fill(Fmt, Args, State) ->
-    ?L2("filling template \"~s\" with~n            ~p", [Fmt, Args], State),
+fill(Fmt, Args) ->
+    ?L2("filling template \"~s\" with~n            ~p", [Fmt, Args]),
     R = rabbit_auth_backend_ldap_util:fill(Fmt, Args),
-    ?L2("template result: \"~s\"", [R], State),
+    ?L2("template result: \"~s\"", [R]),
     R.
 
 log_result({ok, #user{}})   -> ok;
@@ -325,57 +356,3 @@ log_resource(Args) ->
     rabbit_misc:format("~s permission for ~s \"~s\" in \"~s\"",
                        [pget(permission, Args), pget(resource, Args),
                         pget(name, Args), pget(vhost, Args)]).
-
-%%--------------------------------------------------------------------
-
-init([]) ->
-    {ok, list_to_tuple(
-           [state | [get_env(F) || F <- record_info(fields, state)]])}.
-
-handle_call({login, Username}, _From, State) ->
-    %% Without password, e.g. EXTERNAL
-    ?L("CHECK: passwordless login for ~s", [Username], State),
-    R = with_ldap(creds(none, State),
-                  fun(LDAP) -> do_login(Username, none, LDAP, State) end,
-                  State),
-    ?L("DECISION: passwordless login for ~s: ~p",
-       [Username, log_result(R)], State),
-    {reply, R, State};
-
-handle_call({login, Username, Password}, _From, State) ->
-    ?L("CHECK: login for ~s", [Username], State),
-    R = with_ldap({fill_user_dn_pattern(Username, State), Password},
-                  fun(LDAP) -> do_login(Username, Password, LDAP, State) end,
-                  State),
-    ?L("DECISION: login for ~s: ~p", [Username, log_result(R)], State),
-    {reply, R, State};
-
-handle_call({check_vhost, Args, User},
-            _From, State = #state{vhost_access_query = Q}) ->
-    ?L("CHECK: ~s for ~s", [log_vhost(Args), log_user(User)], State),
-    R = evaluate_ldap(Q, Args, User, State),
-    ?L("DECISION: ~s for ~s: ~p",
-       [log_vhost(Args), log_user(User), log_result(R)], State),
-    {reply, R, State};
-
-handle_call({check_resource, Args, User},
-            _From, State = #state{resource_access_query = Q}) ->
-    ?L("CHECK: ~s for ~s", [log_resource(Args), log_user(User)], State),
-    R = evaluate_ldap(Q, Args, User, State),
-    ?L("DECISION: ~s for ~s: ~p",
-       [log_resource(Args), log_user(User), log_result(R)], State),
-    {reply, R, State};
-
-handle_call(_Req, _From, State) ->
-    {reply, unknown_request, State}.
-
-handle_cast(_C, State) ->
-    {noreply, State}.
-
-handle_info(_I, State) ->
-    {noreply, State}.
-
-terminate(_, _) -> ok.
-
-code_change(_, State, _) -> {ok, State}.
-
