@@ -19,6 +19,10 @@
 -behaviour(application).
 -export([start/2, stop/1]).
 
+%% Dummy supervisor to get this application behaviour working
+-behaviour(supervisor).
+-export([init/1]).
+
 start(_Type, _StartArgs) ->
     {ok, Backends} = application:get_env(rabbit, auth_backends),
     case lists:member(rabbit_auth_backend_ldap, Backends) of
@@ -27,7 +31,11 @@ start(_Type, _StartArgs) ->
                    "LDAP plugin loaded, but rabbit_auth_backend_ldap is not "
                    "in the list of auth_backends. LDAP auth will not work.~n")
     end,
-    rabbit_auth_backend_ldap_sup:start_link().
+    supervisor:start_link({local, ?MODULE}, ?MODULE, []).
 
 stop(_State) ->
     ok.
+
+%%----------------------------------------------------------------------------
+
+init([]) -> {ok, {{one_for_one, 3, 10}, []}}.
