@@ -20,7 +20,7 @@
 -include("amqp_client_internal.hrl").
 
 -behaviour(amqp_gen_connection).
--export([init/1, terminate/2, connect/4, do/2, open_channel_args/1, i/2,
+-export([init/0, terminate/2, connect/4, do/2, open_channel_args/1, i/2,
          info_keys/0, handle_message/2, closing/3, channels_terminated/1]).
 
 -define(RABBIT_TCP_OPTS, [binary, {packet, 0}, {active,false}, {nodelay, true}]).
@@ -32,6 +32,7 @@
                 heartbeat,
                 writer0,
                 frame_max,
+                type_sup,
                 closing_reason, %% undefined | Reason
                 waiting_socket_close = false}).
 
@@ -39,7 +40,7 @@
 
 %%---------------------------------------------------------------------------
 
-init([]) ->
+init() ->
     {ok, #state{}}.
 
 open_channel_args(#state{sock = Sock, frame_max = FrameMax}) ->
@@ -107,10 +108,11 @@ info_keys() ->
 %% Handshake
 %%---------------------------------------------------------------------------
 
-connect(AmqpParams = #amqp_params_network{host = Host}, SIF, ChMgr, State) ->
+connect(AmqpParams = #amqp_params_network{host = Host}, SIF, TypeSup, State) ->
     case gethostaddr(Host) of
         []     -> {error, unknown_host};
-        [AF|_] -> do_connect(AF, AmqpParams, SIF, ChMgr, State)
+        [AF|_] -> do_connect(
+                    AF, AmqpParams, SIF, State#state{type_sup = TypeSup})
     end.
 
 do_connect({Addr, Family},
@@ -118,12 +120,12 @@ do_connect({Addr, Family},
                                              port               = Port,
                                              connection_timeout = Timeout,
                                              socket_options     = ExtraOpts},
-           SIF, ChMgr, State) ->
+           SIF, State) ->
     obtain(),
     case gen_tcp:connect(Addr, Port,
                          [Family | ?RABBIT_TCP_OPTS] ++ ExtraOpts,
                          Timeout) of
-        {ok, Sock}     -> try_handshake(AmqpParams, SIF, ChMgr,
+        {ok, Sock}     -> try_handshake(AmqpParams, SIF,
                                         State#state{sock = Sock});
         {error, _} = E -> E
     end;
@@ -132,8 +134,8 @@ do_connect({Addr, Family},
                                              port               = Port,
                                              connection_timeout = Timeout,
                                              socket_options     = ExtraOpts},
-           SIF, ChMgr, State) ->
-    app_utils:start_applications([crypto, public_key, ssl]),
+           SIF, State) ->
+    app_utils:start_applications([asn1, crypto, public_key, ssl]),
     obtain(),
     case gen_tcp:connect(Addr, Port,
                          [Family | ?RABBIT_TCP_OPTS] ++ ExtraOpts,
@@ -142,7 +144,7 @@ do_connect({Addr, Family},
             case ssl:connect(Sock, SslOpts) of
                 {ok, SslSock} ->
                     RabbitSslSock = #ssl_socket{ssl = SslSock, tcp = Sock},
-                    try_handshake(AmqpParams, SIF, ChMgr,
+                    try_handshake(AmqpParams, SIF,
                                   State#state{sock = RabbitSslSock});
                 {error, _} = E ->
                     E
@@ -162,36 +164,40 @@ gethostaddr(Host) ->
                || Family <- inet_address_preference()],
     [{IP, Family} || {Family, {ok, IP}} <- Lookups].
 
-try_handshake(AmqpParams, SIF, ChMgr, State) ->
-    try handshake(AmqpParams, SIF, ChMgr, State) of
+try_handshake(AmqpParams, SIF, State) ->
+    try handshake(AmqpParams, SIF, State) of
         Return -> Return
     catch exit:Reason -> {error, Reason}
     end.
 
-handshake(AmqpParams, SIF, ChMgr, State0 = #state{sock = Sock}) ->
+handshake(AmqpParams, SIF, State0 = #state{sock = Sock}) ->
     ok = rabbit_net:send(Sock, ?PROTOCOL_HEADER),
-    {SHF, State1} = start_infrastructure(SIF, ChMgr, State0),
-    network_handshake(AmqpParams, SHF, State1).
+    network_handshake(AmqpParams, start_infrastructure(SIF, State0)).
 
-start_infrastructure(SIF, ChMgr, State = #state{sock = Sock}) ->
-    {ok, {_MainReader, _AState, Writer, SHF}} = SIF(Sock, ChMgr),
-    {SHF, State#state{writer0 = Writer}}.
+start_infrastructure(SIF, State = #state{sock = Sock}) ->
+    {ok, Writer} = SIF(Sock),
+    State#state{writer0 = Writer}.
 
 network_handshake(AmqpParams = #amqp_params_network{virtual_host = VHost},
-                  SHF, State0) ->
+                  State0) ->
     Start = #'connection.start'{server_properties = ServerProperties,
                                 mechanisms = Mechanisms} =
         handshake_recv('connection.start'),
     ok = check_version(Start),
-    Tune = login(AmqpParams, Mechanisms, State0),
-    {TuneOk, ChannelMax, State1} = tune(Tune, AmqpParams, SHF, State0),
-    do2(TuneOk, State1),
-    do2(#'connection.open'{virtual_host = VHost}, State1),
-    Params = {ServerProperties, ChannelMax, State1},
-    case handshake_recv('connection.open_ok') of
-        #'connection.open_ok'{}                     -> {ok, Params};
-        {closing, #amqp_error{} = AmqpError, Error} -> {closing, Params,
-                                                        AmqpError, Error}
+    case login(AmqpParams, Mechanisms, State0) of
+        {closing, #amqp_error{}, _Error} = Err ->
+            do(#'connection.close_ok'{}, State0),
+            Err;
+        Tune ->
+            {TuneOk, ChannelMax, State1} = tune(Tune, AmqpParams, State0),
+            do2(TuneOk, State1),
+            do2(#'connection.open'{virtual_host = VHost}, State1),
+            Params = {ServerProperties, ChannelMax, State1},
+            case handshake_recv('connection.open_ok') of
+                #'connection.open_ok'{}                     -> {ok, Params};
+                {closing, #amqp_error{} = AmqpError, Error} -> {closing, Params,
+                                                                AmqpError, Error}
+            end
     end.
 
 check_version(#'connection.start'{version_major = ?PROTOCOL_VERSION_MAJOR,
@@ -209,7 +215,7 @@ tune(#'connection.tune'{channel_max = ServerChannelMax,
                         heartbeat   = ServerHeartbeat},
      #amqp_params_network{channel_max = ClientChannelMax,
                           frame_max   = ClientFrameMax,
-                          heartbeat   = ClientHeartbeat}, SHF, State) ->
+                          heartbeat   = ClientHeartbeat}, State) ->
     [ChannelMax, Heartbeat, FrameMax] =
         lists:zipwith(fun (Client, Server) when Client =:= 0; Server =:= 0 ->
                               lists:max([Client, Server]);
@@ -227,17 +233,20 @@ tune(#'connection.tune'{channel_max = ServerChannelMax,
                          _ -> lists:min([FrameMax, ?TCP_MAX_PACKET_SIZE])
                      end,
     NewState = State#state{heartbeat = Heartbeat, frame_max = CappedFrameMax},
-    start_heartbeat(SHF, NewState),
+    start_heartbeat(NewState),
     {#'connection.tune_ok'{channel_max = ChannelMax,
                            frame_max   = CappedFrameMax,
                            heartbeat   = Heartbeat}, ChannelMax, NewState}.
 
-start_heartbeat(SHF, #state{sock = Sock, heartbeat = Heartbeat}) ->
+start_heartbeat(#state{sock      = Sock,
+                       heartbeat = Heartbeat,
+                       type_sup  = Sup}) ->
     Frame = rabbit_binary_generator:build_heartbeat_frame(),
     SendFun = fun () -> catch rabbit_net:send(Sock, Frame) end,
     Connection = self(),
     ReceiveFun = fun () -> Connection ! heartbeat_timeout end,
-    SHF(Sock, Heartbeat, SendFun, Heartbeat, ReceiveFun).
+    rabbit_heartbeat:start(
+      Sup, Sock, Heartbeat, SendFun, Heartbeat, ReceiveFun).
 
 login(Params = #amqp_params_network{auth_mechanisms = ClientMechanisms,
                                     client_properties = UserProps},
@@ -265,7 +274,14 @@ login_loop(Mech, MState0, Params, State) ->
         #'connection.secure'{challenge = Challenge} ->
             {Resp, MState1} = Mech(Challenge, Params, MState0),
             do2(#'connection.secure_ok'{response = Resp}, State),
-            login_loop(Mech, MState1, Params, State)
+            login_loop(Mech, MState1, Params, State);
+        #'connection.close'{reply_code = ?ACCESS_REFUSED,
+                            reply_text = ExplanationBin} ->
+            Explanation = binary_to_list(ExplanationBin),
+            {closing,
+             #amqp_error{name        = access_refused,
+                         explanation = Explanation},
+             {error, {auth_failure, Explanation}}}
     end.
 
 client_properties(UserProperties) ->
@@ -291,6 +307,8 @@ handshake_recv(Expecting) ->
                     Method;
                 {'connection.tune', 'connection.secure'} ->
                     Method;
+                {'connection.tune', 'connection.close'} ->
+                    Method;
                 {'connection.open_ok', _} ->
                     {closing,
                      #amqp_error{name        = command_invalid,
@@ -304,7 +322,7 @@ handshake_recv(Expecting) ->
             end;
         socket_closed ->
             case Expecting of
-                'connection.tune'    -> exit(auth_failure);
+                'connection.tune'    -> exit({auth_failure, "Disconnected"});
                 'connection.open_ok' -> exit(access_refused);
                 _                    -> exit({socket_closed_unexpectedly,
                                               Expecting})
