@@ -19,16 +19,25 @@
 -behaviour(application).
 -export([start/2, stop/1]).
 
-%% This does the same dummy supervisor thing as rabbit_mgmt_app - see the
-%% comment there.
+%% Dummy supervisor - see Ulf Wiger's comment at
+%% http://erlang.2086793.n4.nabble.com/initializing-library-applications-without-processes-td2094473.html
+
+%% All of our actual server processes are supervised by
+%% rabbit_federation_sup, which is started by a rabbit_boot_step
+%% (since it needs to start up before queue / exchange recovery, so it
+%% can't be part of our application).
+%%
+%% However, we still need an application behaviour since we need to
+%% know when our application has started since then the Erlang client
+%% will have started and we can therefore start our links going. Since
+%% the application behaviour needs a tree of processes to supervise,
+%% this is it...
 -behaviour(supervisor).
 -export([init/1]).
 
--include_lib("amqp_client/include/amqp_client.hrl").
--import(rabbit_misc, [pget/3]).
-
 start(_Type, _StartArgs) ->
-    rabbit_federation_link:go(),
+    rabbit_federation_exchange_link:go(),
+    rabbit_federation_queue_link:go(),
     supervisor:start_link({local, ?MODULE}, ?MODULE, []).
 
 stop(_State) ->
