@@ -7,24 +7,31 @@ PROCESS_THRESHOLDS=[[0.75, 'red'],
                     [0.5, 'yellow']];
 
 function fmt_string(str, unknown) {
-    if (unknown == undefined) unkown = UNKNOWN_REPR;
+    if (unknown == undefined) unknown = UNKNOWN_REPR;
     if (str == undefined) return unknown;
     return fmt_escape_html("" + str);
 }
 
 function fmt_bytes(bytes) {
     if (bytes == undefined) return UNKNOWN_REPR;
+    return fmt_si_prefix(bytes, bytes, 1024, false) + 'B';
+}
 
-    function f(n, p) {
-        if (n > 1024) return f(n / 1024, p + 1);
-        else return [n, p];
+function fmt_si_prefix(num0, max0, thousand, allow_fractions) {
+    if (num == 0) return 0;
+
+    function f(n, m, p) {
+        if (m > thousand) return f(n / thousand, m / thousand, p + 1);
+        else return [n, m, p];
     }
 
-    var num_power = f(bytes, 0);
+    var num_power = f(num0, max0, 0);
     var num = num_power[0];
-    var power = num_power[1];
-    var powers = ['B', 'kB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'];
-    return (power == 0 ? num.toFixed(0) : num.toFixed(1)) + powers[power];
+    var max = num_power[1];
+    var power = num_power[2];
+    var powers = ['', 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'];
+    return (((power != 0 || allow_fractions) && max <= 10) ? num.toFixed(1) :
+            num.toFixed(0)) + powers[power];
 }
 
 function fmt_memory(memory, key) {
@@ -182,7 +189,14 @@ function fmt_rate_num(num) {
     if (num == undefined) return UNKNOWN_REPR;
     else if (num < 1)     return num.toFixed(2);
     else if (num < 10)    return num.toFixed(1);
-    else                  return num.toFixed(0);
+    else                  return fmt_num_thousands(num.toFixed(0));
+}
+
+function fmt_num_thousands(num) {
+    if (num == undefined) return UNKNOWN_REPR;
+    num = '' + num;
+    if (num.length < 4) return num;
+    return fmt_num_thousands(num.slice(0, -3)) + ',' + num.slice(-3);
 }
 
 function fmt_rate(obj, name, mode) {
@@ -226,7 +240,8 @@ function fmt_msgs0(obj, name, mode) {
     if (obj == undefined || obj[name] == undefined ||
         obj[name + '_details'] == undefined) return '';
     var details = obj[name + '_details'];
-    return mode == 'avg' ? fmt_rate_num(details.avg) : obj[name];
+    return mode == 'avg' ? fmt_rate_num(details.avg) :
+        fmt_num_thousands(obj[name]);
 }
 
 function fmt_msgs_rate(num) {
@@ -235,15 +250,15 @@ function fmt_msgs_rate(num) {
     else              return '&nbsp;';
 }
 
-function fmt_rate_axis(num) {
-    return num + '/s';
+function fmt_rate_axis(num, max) {
+    return fmt_si_prefix(num, max, 1000, true) + '/s';
 }
 
-function fmt_msgs_axis(num) {
-    return num;
+function fmt_msgs_axis(num, max) {
+    return fmt_si_prefix(num, max, 1000, true);
 }
 
-function fmt_rate_bytes_axis(num) {
+function fmt_rate_bytes_axis(num, max) {
     num = parseInt(num);
     return fmt_bytes(isNaN(num) ? 0 : num) + '/s';
 }
@@ -561,8 +576,12 @@ function fmt_trunc(str, max_length) {
          str.substring(0, max_length) + '...</acronym>') : str;
 }
 
-function alt_rows(i) {
-    return (i % 2 == 0) ? ' class="alt1"' : ' class="alt2"';
+function alt_rows(i, args) {
+    var css = [(i % 2 == 0) ? 'alt1' : 'alt2'];
+    if (args != undefined && args['x-internal-purpose'] != undefined) {
+        css.push('internal-purpose');
+    }
+    return ' class="' + css.join(' ') + '"';
 }
 
 function esc(str) {
@@ -582,13 +601,13 @@ function link_channel(name) {
     return _link_to(short_chan(name), '#/channels/' + esc(name))
 }
 
-function link_exchange(vhost, name) {
+function link_exchange(vhost, name, args) {
     var url = esc(vhost) + '/' + (name == '' ? 'amq.default' : esc(name));
-    return _link_to(fmt_exchange0(name), '#/exchanges/' + url)
+    return _link_to(fmt_exchange0(name), '#/exchanges/' + url, true, args);
 }
 
-function link_queue(vhost, name) {
-    return _link_to(name, '#/queues/' + esc(vhost) + '/' + esc(name))
+function link_queue(vhost, name, args) {
+    return _link_to(name, '#/queues/' + esc(vhost) + '/' + esc(name), true, args);
 }
 
 function link_vhost(name) {
@@ -607,9 +626,15 @@ function link_policy(vhost, name) {
     return _link_to(name, '#/policies/' + esc(vhost) + '/' + esc(name))
 }
 
-function _link_to(name, url, highlight) {
+function _link_to(name, url, highlight, args) {
     if (highlight == undefined) highlight = true;
-    return '<a href="' + url + '">' +
+    var title = null;
+    if (args != undefined && args['x-internal-purpose'] != undefined) {
+        var purpose = args['x-internal-purpose'];
+        title = 'This is used internally by the ' + purpose + ' mechanism.';
+    }
+    return '<a href="' + url + '"' +
+        (title ? ' title="' + title + '"' : '') + '>' +
         (highlight ? fmt_highlight_filter(name) : fmt_escape_html(name)) +
         '</a>';
 }
