@@ -16,15 +16,32 @@ function dispatcher() {
     }
 }
 
+function set_auth_cookie(userinfo) {
+    var b64 = b64_encode_utf8(userinfo);
+    document.cookie = 'auth=' + encodeURIComponent(b64);
+}
+
+function login_route () {
+    var userpass = '' + this.params['username'] + ':' + this.params['password'],
+        location = window.location.href,
+        hash = window.location.hash;
+    set_auth_cookie(decodeURIComponent(userpass));
+    location = location.substr(0, location.length - hash.length);
+    window.location.replace(location);
+    // because we change url, we don't need to hit check_login as
+    // we'll end up doing that at the bottom of start_app_login after
+    // we've changed url.
+}
+
 function start_app_login() {
     app = new Sammy.Application(function () {
         this.put('#/login', function() {
             username = this.params['username'];
             password = this.params['password'];
-            var b64 = b64_encode_utf8(username + ':' + password);
-            document.cookie = 'auth=' + encodeURIComponent(b64);
+            set_auth_cookie(username + ':' + password);
             check_login();
         });
+        this.get('#/login/:username/:password', login_route)
     });
     app.run();
     if (get_cookie('auth') != '') {
@@ -49,7 +66,9 @@ function check_login() {
 }
 
 function start_app() {
-    app.unload();
+    if (app !== undefined) {
+        app.unload();
+    }
     // Oh boy. Sammy uses various different methods to determine if
     // the URL hash has changed. Unsurprisingly this is a native event
     // in modern browsers, and falls back to an icky polling function
@@ -235,7 +254,7 @@ function update_navigation() {
         var val = NAVIGATION[k];
         var path = val;
         while (!leaf(path)) {
-            path = path[keys(path)[0]];
+            path = first_showable_child(path);
         }
         var selected = false;
         if (contains_current_highlight(val)) {
@@ -267,13 +286,23 @@ function nav(pair) {
 }
 
 function show(pair) {
-    return !pair[1] || user_administrator;
+    return jQuery.inArray(pair[1], user_tags) != -1;
 }
 
 function leaf(pair) {
     return typeof(nav(pair)) == 'string';
 }
 
+function first_showable_child(pair) {
+    var items = pair[0];
+    var ks = keys(items);
+    for (var i = 0; i < ks.length; i++) {
+        var child = items[ks[i]];
+        if (show(child)) return child;
+    }
+    return items[ks[0]]; // We'll end up not showing it anyway
+}
+
 function contains_current_highlight(val) {
     if (leaf(val)) {
         return current_highlight == nav(val);
@@ -290,17 +319,19 @@ function contains_current_highlight(val) {
 function obj_to_ul(val) {
     var res = '<ul>';
     for (var k in val) {
-        res += '<li>';
         var obj = val[k];
-        if (leaf(obj) && show(obj)) {
-            res += '<a href="' + nav(obj) + '"' +
-                (current_highlight == nav(obj) ? ' class="selected"' : '') +
-                '>' + k + '</a>';
+        if (show(obj)) {
+            res += '<li>';
+            if (leaf(obj)) {
+                res += '<a href="' + nav(obj) + '"' +
+                    (current_highlight == nav(obj) ? ' class="selected"' : '') +
+                    '>' + k + '</a>';
+            }
+            else {
+                res += obj_to_ul(nav(obj));
+            }
+            res += '</li>';
         }
-        else {
-            res += obj_to_ul(nav(obj));
-        }
-        res += '</li>';
     }
     return res + '</ul>';
 }
@@ -451,7 +482,10 @@ function postprocess() {
             update_manual($(this).attr('for'), $(this).attr('query'));
         });
     $('input, select').die();
-    $('.multifield input').live('blur', function() {
+    $('.multifield input').live('keyup', function() {
+            update_multifields();
+        });
+    $('.multifield select').live('change', function() {
             update_multifields();
         });
     $('.controls-appearance').change(function() {
@@ -523,46 +557,84 @@ function postprocess_partial() {
 }
 
 function update_multifields() {
-    $('.multifield').each(function(index) {
-            var largest_id = 0;
-            var empty_found = false;
-            var name = $(this).attr('id');
-            $('#' + name + ' input[name$="_mfkey"]').each(function(index) {
-                    var match = $(this).attr('name').
-                        match(/[a-z]*_([0-9]*)_mfkey/);
-                    var id = parseInt(match[1]);
-                    largest_id = Math.max(id, largest_id);
-                    var key = $(this).val();
-                    var value = $(this).next('input').val();
-                    if (key == '' && value == '') {
-                        if (empty_found) {
-                            $(this).parent().remove();
-                        }
-                        else {
-                            empty_found = true;
-                        }
-                    }
-                });
-            if (!empty_found) {
-                var prefix = name + '_' + (largest_id + 1);
-                var type_part;
-                if ($(this).hasClass('string-only')) {
-                    type_part = '<input type="hidden" name="' + prefix +
-                        '_mftype" value="string"/>';
-                } else {
-                    type_part = '<select name="' + prefix +
-                        '_mftype">' +
-                        '<option value="string">String</option>' +
-                        '<option value="number">Number</option>' +
-                        '<option value="boolean">Boolean</option>' +
-                        '</select>';
-                }
-                $(this).append('<p><input type="text" name="' + prefix +
-                               '_mfkey" value=""/> = ' +
-                               '<input type="text" name="' + prefix +
-                               '_mfvalue" value=""/> ' + type_part + '</p>');
+    $('div.multifield').each(function(index) {
+        update_multifield($(this), true);
+    });
+}
+
+function update_multifield(multifield, dict) {
+    var largest_id = 0;
+    var empty_found = false;
+    var name = multifield.attr('id');
+    $('#' + name + ' *[name$="_mftype"]').each(function(index) {
+        var re = new RegExp(name + '_([0-9]*)_mftype');
+        var match = $(this).attr('name').match(re);
+        if (!match) return;
+        var id = parseInt(match[1]);
+        largest_id = Math.max(id, largest_id);
+        var prefix = name + '_' + id;
+        var type = $(this).val();
+        var input = $('#' + prefix + '_mfvalue');
+        if (type == 'list') {
+            if (input.size() == 1) {
+                input.replaceWith('<div class="multifield-sub" id="' + prefix +
+                                  '"></div>');
             }
-        });
+            update_multifield($('#' + prefix), false);
+        }
+        else {
+            if (input.size() == 1) {
+                var key = dict ? $('#' + prefix + '_mfkey').val() : '';
+                var value = input.val();
+                if (key == '' && value == '') {
+                    if (empty_found) {
+                        $(this).parent().remove();
+                    }
+                    empty_found = true;
+                }
+            }
+            else {
+                $('#' + prefix).replaceWith(multifield_input(prefix, 'value',
+                                                             'text'));
+            }
+        }
+    });
+    if (!empty_found) {
+        var prefix = name + '_' + (largest_id + 1);
+        var t = multifield.hasClass('string-only') ? 'hidden' : 'select';
+        var val_type = multifield_input(prefix, 'value', 'text') + ' ' +
+            multifield_input(prefix, 'type', t);
+
+        if (dict) {
+            multifield.append('<table><tr><td>' +
+                              multifield_input(prefix, 'key', 'text') +
+                              '</td><td class="equals"> = </td><td>' +
+                              val_type + '</td></tr></table>');
+        }
+        else {
+            multifield.append('<div>' + val_type + '</div>');
+        }
+    }
+}
+
+function multifield_input(prefix, suffix, type) {
+    if (type == 'hidden' ) {
+        return '<input type="hidden" id="' + prefix + '_mf' + suffix +
+            '" name="' + prefix + '_mf' + suffix + '" value="string"/>';
+    }
+    else if (type == 'text' ) {
+        return '<input type="text" id="' + prefix + '_mf' + suffix +
+            '" name="' + prefix + '_mf' + suffix + '" value=""/>';
+    }
+    else if (type == 'select' ) {
+        return '<select id="' + prefix + '_mf' + suffix + '" name="' + prefix +
+            '_mf' + suffix + '">' +
+            '<option value="string">String</option>' +
+            '<option value="number">Number</option>' +
+            '<option value="boolean">Boolean</option>' +
+            '<option value="list">List</option>' +
+            '</select>';
+    }
 }
 
 function update_filter() {
@@ -689,9 +761,16 @@ function replace_content(id, html) {
     $("#" + id).html(html);
 }
 
+var ejs_cached = {};
+
 function format(template, json) {
     try {
-        var tmpl = new EJS({url: 'js/tmpl/' + template + '.ejs'});
+        var cache = true;
+        if (!(template in ejs_cached)) {
+            ejs_cached[template] = true;
+            cache = false;
+        }
+        var tmpl = new EJS({url: 'js/tmpl/' + template + '.ejs', cache: cache});
         return tmpl.render(json);
     } catch (err) {
         clearInterval(timer);
@@ -855,11 +934,16 @@ function params_magic(params) {
 }
 
 function collapse_multifields(params0) {
+    function set(x) { return x != '' && x != undefined }
+
     var params = {};
-    for (key in params0) {
-        var match = key.match(/([a-z]*)_([0-9]*)_mfkey/);
-        var match2 = key.match(/[a-z]*_[0-9]*_mfvalue/);
-        var match3 = key.match(/[a-z]*_[0-9]*_mftype/);
+    var ks = keys(params0);
+    var ids = [];
+    for (i in ks) {
+        var key = ks[i];
+        var match = key.match(/([a-z]*)_([0-9_]*)_mftype/);
+        var match2 = key.match(/[a-z]*_[0-9_]*_mfkey/);
+        var match3 = key.match(/[a-z]*_[0-9_]*_mfvalue/);
         if (match == null && match2 == null && match3 == null) {
             params[key] = params0[key];
         }
@@ -869,27 +953,50 @@ function collapse_multifields(params0) {
         else {
             var name = match[1];
             var id = match[2];
-            if (params[name] == undefined) {
-                params[name] = {};
+            ids.push([name, id]);
+        }
+    }
+    ids.sort();
+    var id_map = {};
+    for (i in ids) {
+        var name = ids[i][0];
+        var id = ids[i][1];
+        if (params[name] == undefined) {
+            params[name] = {};
+            id_map[name] = {};
+        }
+        var id_parts = id.split('_');
+        var k = params0[name + '_' + id_parts[0] + '_mfkey'];
+        var v = params0[name + '_' + id + '_mfvalue'];
+        var t = params0[name + '_' + id + '_mftype'];
+        var val = null;
+        if (t == 'list') {
+            val = [];
+            id_map[name][id] = val;
+        }
+        else if (set(k) || set(v)) {
+            if (t == 'boolean') {
+                if (v != 'true' && v != 'false')
+                    throw(k + ' must be "true" or "false"; got ' + v);
+                val = (v == 'true');
             }
-            if (params0[key] != "") {
-                var k = params0[key];
-                var v = params0[name + '_' + id + '_mfvalue'];
-                var t = params0[name + '_' + id + '_mftype'];
-                if (t == 'boolean') {
-                    if (v != 'true' && v != 'false')
-                        throw(k + ' must be "true" or "false"; got ' + v);
-                    params[name][k] = (v == 'true');
-                }
-                else if (t == 'number') {
-                    var n = parseFloat(v);
-                    if (isNaN(n))
-                        throw(k + ' must be a number; got ' + v);
-                    params[name][k] = n;
-                }
-                else {
-                    params[name][k] = v;
-                }
+            else if (t == 'number') {
+                var n = parseFloat(v);
+                if (isNaN(n))
+                    throw(k + ' must be a number; got ' + v);
+                val = n;
+            }
+            else {
+                val = v;
+            }
+        }
+        if (val != null) {
+            if (id_parts.length == 1) {
+                params[name][k] = val;
+            }
+            else {
+                var prefix = id_parts.slice(0, id_parts.length - 1).join('_');
+                id_map[name][prefix].push(val);
             }
         }
     }
@@ -907,9 +1014,6 @@ function add_known_arguments(params) {
                     throw(k + " must be an integer.");
                 }
             }
-            else if (type == 'array' && typeof(v) == 'string') {
-                v = v.split(' ');
-            }
             params.arguments[k] = v;
         }
         delete params[k];
