@@ -14,7 +14,7 @@ function clear_pref(k) {
 
 function get_pref(k) {
     var r = parse_cookie()[short_key(k)];
-    return r == undefined ? null : r;
+    return r == undefined ? default_pref(k) : r;
 }
 
 function section_pref(template, name) {
@@ -23,6 +23,16 @@ function section_pref(template, name) {
 
 // ---------------------------------------------------------------------------
 
+function default_pref(k) {
+    if (k.substring(0, 12) == 'chart-range-') return '60|5';
+    if (k.substring(0, 11) == 'chart-size-')  return 'small';
+    if (k.substring(0, 10) == 'rate-mode-')   return 'chart';
+    if (k == 'truncate')                      return '100';
+    return null;
+}
+
+// ---------------------------------------------------------------------------
+
 function parse_cookie() {
     var c = get_cookie('m');
     var items = c.length == 0 ? [] : c.split('|');
