@@ -179,19 +179,14 @@ prioritise_cast(_Msg, _Len, _State) ->
 %%----------------------------------------------------------------------------
 
 start_link() ->
-    %% When failing over it is possible that the mirrored_supervisor
-    %% might hear of the death of the old DB, and start a new one,
-    %% before the global name server notices. Therefore rather than
-    %% telling gen_server:start_link/4 to register it for us, we
-    %% invoke global:re_register_name/2 ourselves, and just steal the
-    %% name if it existed before. We therefore rely on
-    %% mirrored_supervisor to maintain the uniqueness of this process.
-    case gen_server2:start_link(?MODULE, [], []) of
-        {ok, Pid} -> yes = global:re_register_name(?MODULE, Pid),
+    case gen_server2:start_link({global, ?MODULE}, ?MODULE, [], []) of
+        {ok, Pid} -> register(?MODULE, Pid), %% [1]
                      rabbit:force_event_refresh(),
                      {ok, Pid};
         Else      -> Else
     end.
+%% [1] For debugging it's helpful to locally register the name too
+%% since that shows up in places global names don't.
 
 %% R = Ranges, M = Mode
 augment_exchanges(Xs, R, M) -> safe_call({augment_exchanges, Xs, R, M}, Xs).
@@ -211,12 +206,19 @@ get_overview(R)             -> safe_call({get_overview, all, R}).
 override_lookups(Lookups)   -> safe_call({override_lookups, Lookups}).
 reset_lookups()             -> safe_call(reset_lookups).
 
-safe_call(Term) -> safe_call(Term, []).
+safe_call(Term)          -> safe_call(Term, []).
+safe_call(Term, Default) -> safe_call(Term, Default, 1).
 
-safe_call(Term, Item) ->
+%% See rabbit_mgmt_sup_sup for a discussion of the retry logic.
+safe_call(Term, Default, Retries) ->
     try
         gen_server2:call({global, ?MODULE}, Term, infinity)
-    catch exit:{noproc, _} -> Item
+    catch exit:{noproc, _} ->
+            case Retries of
+                0 -> Default;
+                _ -> rabbit_mgmt_sup_sup:start_child(),
+                     safe_call(Term, Default, Retries - 1)
+            end
     end.
 
 %%----------------------------------------------------------------------------
@@ -420,14 +422,10 @@ lookup_element(Table, Key, Pos) ->
 fine_stats_id(ChPid, {Q, X}) -> {ChPid, Q, X};
 fine_stats_id(ChPid, QorX)   -> {ChPid, QorX}.
 
-floor(TS, State) -> floor0(rabbit_mgmt_format:timestamp_ms(TS), State).
-ceil (TS, State) -> ceil0 (rabbit_mgmt_format:timestamp_ms(TS), State).
-
-floor0(MS, #state{interval = Interval})        -> (MS div Interval) * Interval.
-ceil0(MS, State = #state{interval = Interval}) -> case floor0(MS, State) of
-                                                      MS    -> MS;
-                                                      Floor -> Floor + Interval
-                                                  end.
+floor(TS, #state{interval = Interval}) ->
+    rabbit_mgmt_util:floor(rabbit_mgmt_format:timestamp_ms(TS), Interval).
+ceil(TS, #state{interval = Interval}) ->
+    rabbit_mgmt_util:ceil (rabbit_mgmt_format:timestamp_ms(TS), Interval).
 
 details_key(Key) -> list_to_atom(atom_to_list(Key) ++ "_details").
 
@@ -523,21 +521,17 @@ handle_event(Event = #event{type = channel_closed,
     ets:match_delete(Old, {{fine, {Pid, '_', '_'}}, '_'});
 
 handle_event(#event{type = consumer_created, props = Props}, State) ->
-    handle_consumer(fun(Table, Id, P) -> ets:insert(Table, {Id, P}) end,
+    Fmt = [{fun rabbit_mgmt_format:amqp_table/1, [arguments]}],
+    handle_consumer(fun(Table, Id, P0) ->
+                            P = rabbit_mgmt_format:format(P0, Fmt),
+                            ets:insert(Table, {Id, P})
+                    end,
                     Props, State);
 
 handle_event(#event{type = consumer_deleted, props = Props}, State) ->
     handle_consumer(fun(Table, Id, _P) -> ets:delete(Table, Id) end,
                     Props, State);
 
-handle_event(#event{type = queue_mirror_deaths, props = Props},
-             #state{tables = Tables}) ->
-    Dead = pget(pids, Props),
-    Table = orddict:fetch(queue_stats, Tables),
-    %% Only the master can be in the DB, but it's easier just to
-    %% delete all of them
-    [ets:delete(Table, {Pid, stats}) || Pid <- Dead];
-
 %% TODO: we don't clear up after dead nodes here - this is a very tiny
 %% leak every time a node is permanently removed from the cluster. Do
 %% we care?
@@ -920,7 +914,8 @@ format_detail_id(#resource{name = Name, virtual_host = Vhost, kind = Kind},
 
 format_samples(Ranges, ManyStats, #state{interval = Interval}) ->
     lists:append(
-      [case rabbit_mgmt_stats:is_blank(Stats) of
+      [case rabbit_mgmt_stats:is_blank(Stats) andalso
+           not lists:member(K, ?COARSE_QUEUE_STATS) of
            true  -> [];
            false -> {Details, Counter} = rabbit_mgmt_stats:format(
                                            pick_range(K, Ranges),
@@ -945,8 +940,10 @@ pick_range(K, {RangeL, RangeM, RangeD}) ->
 adjust_hibernated_memory_use(Qs) ->
     Pids = [pget(pid, Q) ||
                Q <- Qs, pget(idle_since, Q, not_idle) =/= not_idle],
-    {Mem, _BadNodes} = delegate:invoke(
-                         Pids, fun (Pid) -> process_info(Pid, memory) end),
+    %% We use delegate here not for ordering reasons but because we
+    %% want to get the right amount of parallelism and minimise
+    %% cross-cluster communication.
+    {Mem, _BadNodes} = delegate:invoke(Pids, {erlang, process_info, [memory]}),
     MemDict = dict:from_list([{P, M} || {P, M = {memory, _}} <- Mem]),
     [case dict:find(pget(pid, Q), MemDict) of
          error        -> Q;
