@@ -10,8 +10,8 @@
 %%
 %%   The Original Code is RabbitMQ Management Plugin.
 %%
-%%   The Initial Developer of the Original Code is VMware, Inc.
-%%   Copyright (c) 2010-2013 VMware, Inc.  All rights reserved.
+%%   The Initial Developer of the Original Code is GoPivotal, Inc.
+%%   Copyright (c) 2010-2013 GoPivotal, Inc.  All rights reserved.
 %%
 
 -module(rabbit_mgmt_load_definitions).
@@ -39,10 +39,10 @@ maybe_load_definitions() ->
                     {ok, Body} -> rabbit_log:info(
                                     "Applying definitions from: ~s~n", [File]),
                                   load_definitions(Body);
-                    {error, E} -> exit({could_not_read_defs, {File, E}})
+                    {error, E} -> {error, {could_not_read_defs, {File, E}}}
                 end
     end.
 
 load_definitions(Body) ->
     rabbit_mgmt_wm_definitions:apply_defs(
-      Body, fun rabbit_misc:const_ok/0, fun (E) -> exit(E) end).
+      Body, fun rabbit_misc:const_ok/0, fun (E) -> {error, E} end).
