@@ -24,11 +24,11 @@
 -include_lib("rabbit_common/include/rabbit.hrl").
 
 init([]) ->
-    {ok, DBModule} = application:get_env(rabbitmq_management, db_module),
-    DB = {rabbit_mgmt_db,
-              {DBModule, start_link, []},
-              permanent, ?MAX_WAIT, worker, [DBModule]},
+    DB = {rabbit_mgmt_db, {rabbit_mgmt_db, start_link, []},
+          permanent, ?MAX_WAIT, worker, [rabbit_mgmt_db]},
     {ok, {{one_for_one, 10, 10}, [DB]}}.
 
 start_link() ->
-     mirrored_supervisor:start_link({local, ?MODULE}, ?MODULE, ?MODULE, []).
+     mirrored_supervisor:start_link(
+       {local, ?MODULE}, ?MODULE, fun rabbit_misc:execute_mnesia_transaction/1,
+       ?MODULE, []).
