@@ -21,7 +21,7 @@
 -export([is_authorized/2, is_authorized_admin/2, is_authorized_admin/4,
          vhost/1]).
 -export([is_authorized_vhost/2, is_authorized_user/3,
-         is_authorized_monitor/2]).
+         is_authorized_monitor/2, is_authorized_policies/2]).
 -export([bad_request/3, bad_request_exception/4, id/2, parse_bool/1,
          parse_int/1]).
 -export([with_decode/4, not_found/3, amqp_request/4]).
@@ -32,7 +32,8 @@
 -export([with_decode/5, decode/1, decode/2, redirect/2, args/1]).
 -export([reply_list/3, reply_list/4, sort_list/2, destination_type/1]).
 -export([post_respond/1, columns/1, is_monitor/1]).
--export([list_visible_vhosts/1, b64decode_or_throw/1, no_range/0, range/1]).
+-export([list_visible_vhosts/1, b64decode_or_throw/1, no_range/0, range/1,
+         range_ceil/1, floor/2, ceil/2]).
 
 -import(rabbit_misc, [pget/2, pget/3]).
 
@@ -65,14 +66,16 @@ is_authorized_vhost(ReqData, Context) ->
     is_authorized(ReqData, Context,
                   <<"User not authorised to access virtual host">>,
                   fun(User) ->
-                          case vhost(ReqData) of
-                              not_found -> true;
-                              none      -> true;
-                              V         -> lists:member(
-                                             V, list_login_vhosts(User))
-                          end
+                          user_matches_vhost(ReqData, User)
                   end).
 
+user_matches_vhost(ReqData, User) ->
+    case vhost(ReqData) of
+        not_found -> true;
+        none      -> true;
+        V         -> lists:member(V, list_login_vhosts(User))
+    end.
+
 %% Used for connections / channels. A normal user can only see / delete
 %% their own stuff. Monitors can see other users' and delete their
 %% own. Admins can do it all.
@@ -86,6 +89,16 @@ is_authorized_user(ReqData, Context, Item) ->
                           end orelse Username == pget(user, Item)
                   end).
 
+%% For policies / parameters. Like is_authorized_vhost but you have to
+%% be a policymaker.
+is_authorized_policies(ReqData, Context) ->
+    is_authorized(ReqData, Context,
+                  <<"User not authorised to access object">>,
+                  fun(User = #user{tags = Tags}) ->
+                          is_policymaker(Tags) andalso
+                              user_matches_vhost(ReqData, User)
+                  end).
+
 is_authorized(ReqData, Context, ErrorMsg, Fun) ->
     case rabbit_web_dispatch_util:parse_auth_header(
            wrq:get_req_header("authorization", ReqData)) of
@@ -286,6 +299,9 @@ decode(Keys, Body) ->
         Else     -> Else
     end.
 
+decode(<<"">>) ->
+    {ok, []};
+
 decode(Body) ->
     try
         {struct, J} = mochijson2:decode(Body),
@@ -348,6 +364,7 @@ parse_bool(<<"true">>)  -> true;
 parse_bool(<<"false">>) -> false;
 parse_bool(true)        -> true;
 parse_bool(false)       -> false;
+parse_bool(undefined)   -> undefined;
 parse_bool(V)           -> throw({error, {not_boolean, V}}).
 
 parse_int(I) when is_integer(I) -> I;
@@ -404,8 +421,8 @@ with_channel(VHost, ReqData,
             catch amqp_channel:close(Ch),
             catch amqp_connection:close(Conn)
             end;
-        {error, auth_failure} ->
-            not_authorised(<<"">>, ReqData, Context);
+        {error, {auth_failure, Msg}} ->
+            not_authorised(Msg, ReqData, Context);
         {error, {nodedown, N}} ->
             bad_request(
               list_to_binary(
@@ -463,9 +480,11 @@ post_respond({JSON, ReqData, Context}) ->
              "content-type", "application/json",
              wrq:append_to_response_body(JSON, ReqData)), Context}.
 
-is_admin(T)     -> intersects(T, [administrator]).
-is_monitor(T)   -> intersects(T, [administrator, monitoring]).
-is_mgmt_user(T) -> intersects(T, [administrator, monitoring, management]).
+is_admin(T)       -> intersects(T, [administrator]).
+is_policymaker(T) -> intersects(T, [administrator, policymaker]).
+is_monitor(T)     -> intersects(T, [administrator, monitoring]).
+is_mgmt_user(T)   -> intersects(T, [administrator, monitoring, policymaker,
+                                    management]).
 
 intersects(A, B) -> lists:any(fun(I) -> lists:member(I, B) end, A).
 
@@ -501,22 +520,33 @@ b64decode_or_throw(B64) ->
 
 no_range() -> {no_range, no_range, no_range}.
 
-range(ReqData) -> {range("lengths",    ReqData),
-                   range("msg_rates",  ReqData),
-                   range("data_rates", ReqData)}.
+%% Take floor on queries so we make sure we only return samples
+%% for which we've finished receiving events. Fixes the "drop at
+%% the end" problem.
+range(ReqData) -> {range("lengths",    fun floor/2, ReqData),
+                   range("msg_rates",  fun floor/2, ReqData),
+                   range("data_rates", fun floor/2, ReqData)}.
 
-range(Prefix, ReqData) ->
+%% ...but if we know only one event could have contributed towards
+%% what we are interested in, then let's take the ceiling instead and
+%% get slightly fresher data.
+%%
+%% Why does msg_rates still use floor/2? Because in the cases where we
+%% call this function (for connections and queues) the msg_rates are still
+%% aggregated even though the lengths and data rates aren't.
+range_ceil(ReqData) -> {range("lengths",    fun ceil/2,  ReqData),
+                        range("msg_rates",  fun floor/2, ReqData),
+                        range("data_rates", fun ceil/2,  ReqData)}.
+
+range(Prefix, Round, ReqData) ->
     Age0 = int(Prefix ++ "_age", ReqData),
     Incr0 = int(Prefix ++ "_incr", ReqData),
     if
         is_integer(Age0) andalso is_integer(Incr0) ->
             Age = Age0 * 1000,
             Incr = Incr0 * 1000,
-            %% Take floor on queries so we make sure we only return samples
-            %% for which we've finished receiving events. Fixes the "drop at
-            %% the end" problem.
             Now = rabbit_mgmt_format:timestamp_ms(erlang:now()),
-            Last = (Now div Incr) * Incr,
+            Last = Round(Now, Incr),
             #range{first = (Last - Age),
                    last  = Last,
                    incr  = Incr};
@@ -524,6 +554,13 @@ range(Prefix, ReqData) ->
             no_range
     end.
 
+floor(TS, Interval) -> (TS div Interval) * Interval.
+
+ceil(TS, Interval) -> case floor(TS, Interval) of
+                          TS    -> TS;
+                          Floor -> Floor + Interval
+                      end.
+
 int(Name, ReqData) ->
     case wrq:get_qs_value(Name, ReqData) of
         undefined -> undefined;
