@@ -63,4 +63,4 @@ is_authorized(ReqData, Context) ->
 
 conn(ReqData) ->
     rabbit_mgmt_db:get_connection(rabbit_mgmt_util:id(connection, ReqData),
-                                  rabbit_mgmt_util:range(ReqData)).
+                                  rabbit_mgmt_util:range_ceil(ReqData)).
