@@ -22,7 +22,7 @@
 
 -export([apply_defs/3]).
 
--import(rabbit_misc, [pget/2]).
+-import(rabbit_misc, [pget/2, pget/3]).
 
 -include("rabbit_mgmt.hrl").
 -include_lib("webmachine/include/webmachine.hrl").
@@ -140,6 +140,7 @@ apply_defs(Body, SuccessFun, ErrorFun) ->
     end.
 
 get_part(Name, Parts) ->
+    %% TODO any reason not to use lists:keyfind instead?
     Filtered = [Value || {N, _Meta, Value} <- Parts, N == Name],
     case Filtered of
         []  -> unknown;
@@ -150,10 +151,10 @@ export_queue(Queue) ->
     pget(owner_pid, Queue) == none.
 
 export_binding(Binding, Qs) ->
-    Src      = pget(source, Binding),
-    Dest     = pget(destination, Binding),
+    Src      = pget(source,           Binding),
+    Dest     = pget(destination,      Binding),
     DestType = pget(destination_type, Binding),
-    VHost    = pget(vhost, Binding),
+    VHost    = pget(vhost,            Binding),
     Src =/= <<"">>
         andalso
           ( (DestType =:= queue andalso lists:member({Dest, VHost}, Qs))
@@ -173,7 +174,7 @@ rw_state() ->
      {vhosts,      [name]},
      {permissions, [user, vhost, configure, write, read]},
      {parameters,  [vhost, component, name, value]},
-     {policies,    [vhost, name, pattern, definition, priority]},
+     {policies,    [vhost, name, pattern, definition, priority, 'apply-to']},
      {queues,      [name, vhost, durable, auto_delete, arguments]},
      {exchanges,   [name, vhost, type, durable, auto_delete, internal,
                     arguments]},
@@ -193,22 +194,19 @@ filter_item(Item, Allowed) ->
 
 for_all(Name, All, Fun) ->
     case pget(Name, All) of
-        undefined ->
-            ok;
-        List ->
-            [Fun([{atomise_name(K), V} || {K, V} <- I]) ||
-                {struct, I} <- List]
+        undefined -> ok;
+        List      -> [Fun([{atomise_name(K), V} || {K, V} <- I]) ||
+                         {struct, I} <- List]
     end.
 
-atomise_name(N) ->
-    list_to_atom(binary_to_list(N)).
+atomise_name(N) -> list_to_atom(binary_to_list(N)).
 
 %%--------------------------------------------------------------------
 
 add_parameter(Param) ->
-    VHost = pget(vhost, Param),
-    Comp = pget(component, Param),
-    Key = pget(name, Param),
+    VHost = pget(vhost,     Param),
+    Comp  = pget(component, Param),
+    Key   = pget(name,      Param),
     case rabbit_runtime_parameters:set(
            VHost, Comp, Key, rabbit_misc:json_to_term(pget(value, Param))) of
         ok                -> ok;
@@ -219,11 +217,12 @@ add_parameter(Param) ->
 
 add_policy(Param) ->
     VHost = pget(vhost, Param),
-    Key = pget(name, Param),
+    Key   = pget(name,  Param),
     case rabbit_policy:set(
            VHost, Key, pget(pattern, Param),
            rabbit_misc:json_to_term(pget(definition, Param)),
-           pget(priority, Param)) of
+           pget(priority, Param),
+           pget('apply-to', Param, <<"all">>)) of
         ok                -> ok;
         {error_string, E} -> S = rabbit_misc:format(" (~s/~s)", [VHost, Key]),
                              exit(list_to_binary(E ++ S))
@@ -234,7 +233,8 @@ add_user(User) ->
 
 add_vhost(VHost) ->
     VHostName = pget(name, VHost),
-    rabbit_mgmt_wm_vhost:put_vhost(VHostName).
+    VHostTrace = pget(tracing, VHost),
+    rabbit_mgmt_wm_vhost:put_vhost(VHostName, VHostTrace).
 
 add_permission(Permission) ->
     rabbit_auth_backend_internal:set_permissions(pget(user,      Permission),
@@ -270,8 +270,7 @@ add_binding(Binding) ->
                key          = pget(routing_key,                     Binding),
                args         = rabbit_mgmt_util:args(pget(arguments, Binding))}).
 
-r(Type, Props) ->
-    r(Type, name, Props).
+r(Type, Props) -> r(Type, name, Props).
 
 r(Type, Name, Props) ->
     rabbit_misc:r(pget(vhost, Props), Type, pget(Name, Props)).
