@@ -71,7 +71,6 @@ process_request(?CONNECT,
                         case process_login(UserBin, PassBin, PState) of
                             {?CONNACK_ACCEPT, Conn} ->
                                 link(Conn),
-                                maybe_clean_sess(CleanSess, Conn, ClientId),
                                 {ok, Ch} = amqp_connection:open_channel(Conn),
                                 ok = rabbit_mqtt_collector:register(
                                   ClientId, self()),
@@ -79,11 +78,12 @@ process_request(?CONNECT,
                                 #'basic.qos_ok'{} = amqp_channel:call(
                                   Ch, #'basic.qos'{prefetch_count = Prefetch}),
                                 {?CONNACK_ACCEPT,
-                                 PState #proc_state{ will_msg   = make_will_msg(Var),
-                                                     clean_sess = CleanSess,
-                                                     channels   = {Ch, undefined},
-                                                     connection = Conn,
-                                                     client_id  = ClientId }};
+                                 maybe_clean_sess(
+                                   PState #proc_state{ will_msg   = make_will_msg(Var),
+                                                       clean_sess = CleanSess,
+                                                       channels   = {Ch, undefined},
+                                                       connection = Conn,
+                                                       client_id  = ClientId })};
                             ConnAck ->
                                 {ConnAck, PState}
                         end
@@ -245,7 +245,8 @@ amqp_callback({#'basic.deliver'{ consumer_tag = ConsumerTag,
 
 amqp_callback(#'basic.ack'{ multiple = true, delivery_tag = Tag } = Ack,
               PState = #proc_state{ unacked_pubs = UnackedPubs }) ->
-    case gb_trees:take_smallest(UnackedPubs) of
+    case gb_trees:size(UnackedPubs) > 0 andalso
+         gb_trees:take_smallest(UnackedPubs) of
         {TagSmall, MsgId, UnackedPubs1} when TagSmall =< Tag ->
             send_client(
               #mqtt_frame{ fixed    = #mqtt_frame_fixed{ type = ?PUBACK },
@@ -288,17 +289,20 @@ delivery_qos(Tag, Headers,   #proc_state{ consumer_tags = {_, Tag} }) ->
         undefined   -> {?QOS_1, ?QOS_1}
     end.
 
-maybe_clean_sess(false, _Conn, _ClientId) ->
-    % todo: establish subscription to deliver old unacknowledged messages
-    ok;
-maybe_clean_sess(true, Conn, ClientId) ->
+maybe_clean_sess(PState = #proc_state { clean_sess = false }) ->
+    {_Queue, PState1} = ensure_queue(?QOS_1, PState),
+    PState1;
+maybe_clean_sess(PState = #proc_state { clean_sess = true,
+                                        connection = Conn,
+                                        client_id  = ClientId }) ->
     {_, Queue} = rabbit_mqtt_util:subcription_queue_name(ClientId),
     {ok, Channel} = amqp_connection:open_channel(Conn),
     try amqp_channel:call(Channel, #'queue.delete'{ queue = Queue }) of
         #'queue.delete_ok'{} -> ok = amqp_channel:close(Channel)
     catch
         exit:_Error -> ok
-    end.
+    end,
+    PState.
 
 %%----------------------------------------------------------------------------
 
@@ -322,16 +326,17 @@ process_login(UserBin, PassBin, #proc_state{ channels  = {undefined, undefined},
                                   password     = PassBin,
                                   virtual_host = VHost,
                                   adapter_info = adapter_info(Sock)}) of
-         {ok, Connection}        -> {?CONNACK_ACCEPT, Connection};
-         {error, auth_failure}   -> rabbit_log:error(
-                                      "MQTT login failed for ~p auth_failure~n",
-                                      [binary_to_list(UserBin)]),
-                                    ?CONNACK_CREDENTIALS;
-         {error, access_refused} -> rabbit_log:warning(
-                                      "MQTT login failed for ~p access_refused "
-                                      "(vhost access not allowed)~n",
-                                      [binary_to_list(UserBin)]),
-                                    ?CONNACK_AUTH
+         {ok, Connection} ->
+             {?CONNACK_ACCEPT, Connection};
+         {error, {auth_failure, Explanation}} ->
+             rabbit_log:error("MQTT login failed for ~p auth_failure: ~s~n",
+                              [binary_to_list(UserBin), Explanation]),
+             ?CONNACK_CREDENTIALS;
+         {error, access_refused} ->
+             rabbit_log:warning("MQTT login failed for ~p access_refused "
+                                "(vhost access not allowed)~n",
+                                [binary_to_list(UserBin)]),
+             ?CONNACK_AUTH
       end.
 
 creds(User, Pass) ->
@@ -358,6 +363,9 @@ supported_subs_qos(?QOS_0) -> ?QOS_0;
 supported_subs_qos(?QOS_1) -> ?QOS_1;
 supported_subs_qos(?QOS_2) -> ?QOS_1.
 
+delivery_mode(?QOS_0) -> 1;
+delivery_mode(?QOS_1) -> 2.
+
 %% different qos subscriptions are received in different queues
 %% with appropriate durability and timeout arguments
 %% this will lead to duplicate messages for overlapping subscriptions
@@ -439,7 +447,8 @@ amqp_pub(#mqtt_msg{ qos        = Qos,
                                    rabbit_mqtt_util:mqtt2amqp(Topic)},
     Headers = [{<<"x-mqtt-publish-qos">>, byte, Qos},
                {<<"x-mqtt-dup">>, bool, Dup}],
-    Msg = #amqp_msg{ props   = #'P_basic'{ headers = Headers },
+    Msg = #amqp_msg{ props   = #'P_basic'{ headers       = Headers,
+                                           delivery_mode = delivery_mode(Qos)},
                      payload = Payload },
     {UnackedPubs1, Ch, SeqNo1} =
         case Qos =:= ?QOS_1 andalso MessageId =/= undefined of
@@ -461,7 +470,13 @@ send_client(Frame, #proc_state{ socket = Sock }) ->
 
 close_connection(PState = #proc_state{ connection = undefined }) ->
     PState;
-close_connection(PState = #proc_state{ connection = Connection }) ->
+close_connection(PState = #proc_state{ connection = Connection,
+                                       client_id  = ClientId }) ->
+    % todo: maybe clean session
+    case ClientId of
+        undefined -> ok;
+        _         -> ok = rabbit_mqtt_collector:unregister(ClientId, self())
+    end,
     %% ignore noproc or other exceptions to avoid debris
     catch amqp_connection:close(Connection),
     PState #proc_state{ channels   = {undefined, undefined},
