@@ -17,7 +17,7 @@
 -module(rabbit_stomp_processor).
 -behaviour(gen_server2).
 
--export([start_link/1, process_frame/2, flush_and_die/1]).
+-export([start_link/1, init_arg/2, process_frame/2, flush_and_die/1]).
 -export([init/1, handle_call/3, handle_cast/2, handle_info/2,
          code_change/3, terminate/2]).
 
@@ -42,6 +42,9 @@
 start_link(Args) ->
     gen_server2:start_link(?MODULE, Args, []).
 
+init_arg(ProcessorPid, InitArgs) ->
+    gen_server2:cast(ProcessorPid, {init, InitArgs}).
+
 process_frame(Pid, Frame = #stomp_frame{command = "SEND"}) ->
     credit_flow:send(Pid),
     gen_server2:cast(Pid, {"SEND", Frame, self()});
@@ -55,7 +58,7 @@ flush_and_die(Pid) ->
 %% Basic gen_server2 callbacks
 %%----------------------------------------------------------------------------
 
-init([SendFun, AdapterInfo, StartHeartbeatFun, SSLLoginName, Configuration]) ->
+init(Configuration) ->
     process_flag(trap_exit, true),
     {ok,
      #state {
@@ -64,15 +67,11 @@ init([SendFun, AdapterInfo, StartHeartbeatFun, SSLLoginName, Configuration]) ->
        connection          = none,
        subscriptions       = dict:new(),
        version             = none,
-       start_heartbeat_fun = StartHeartbeatFun,
        pending_receipts    = undefined,
        config              = Configuration,
        route_state         = rabbit_routing_util:init_state(),
        reply_queues        = dict:new(),
-       frame_transformer   = undefined,
-       adapter_info        = AdapterInfo,
-       send_fun            = SendFun,
-       ssl_login_name      = SSLLoginName},
+       frame_transformer   = undefined},
      hibernate,
      {backoff, 1000, 1000, 10000}
     }.
@@ -80,6 +79,13 @@ init([SendFun, AdapterInfo, StartHeartbeatFun, SSLLoginName, Configuration]) ->
 terminate(_Reason, State) ->
     close_connection(State).
 
+handle_cast({init, [SendFun, AdapterInfo, StartHeartbeatFun, SSLLoginName]},
+            State) ->
+    {noreply, State #state { send_fun            = SendFun,
+                             adapter_info        = AdapterInfo,
+                             start_heartbeat_fun = StartHeartbeatFun,
+                             ssl_login_name      = SSLLoginName }};
+
 handle_cast(flush_and_die, State) ->
     {stop, normal, close_connection(State)};
 
@@ -197,8 +203,8 @@ process_connect(Implicit, Frame,
                       FT = frame_transformer(Version),
                       Frame1 = FT(Frame),
                       {Username, Passwd} = creds(Frame1, SSLLoginName, Config),
-                      {ok, DefaultVHost} =
-                          application:get_env(rabbit, default_vhost),
+                      {ok, DefaultVHost} = application:get_env(
+                                             rabbitmq_stomp, default_vhost),
                       {ProtoName, _} = AdapterInfo#amqp_adapter_info.protocol,
                       Res = do_login(
                               Username, Passwd,
@@ -516,7 +522,7 @@ do_login(Username, Passwd, VirtualHost, Heartbeat, AdapterInfo, Version,
                             channel    = Channel,
                             connection = Connection,
                             version    = Version});
-        {error, auth_failure} ->
+        {error, {auth_failure, _}} ->
             rabbit_log:warning("STOMP login failed for user ~p~n",
                                [binary_to_list(Username)]),
             error("Bad CONNECT", "Access refused for user '" ++
@@ -678,6 +684,8 @@ send_method(Method = #'basic.publish'{}, Channel, Properties, BodyFragments,
                 payload = list_to_binary(BodyFragments)}),
     State.
 
+close_connection(State = #state{connection = none}) ->
+    State;
 %% Closing the connection will close the channel and subchannels
 close_connection(State = #state{connection = Connection}) ->
     %% ignore noproc or other exceptions to avoid debris
