@@ -347,19 +347,25 @@ handle_app_error(App, Reason) ->
 
 start_it(StartFun) ->
     Marker = spawn_link(fun() -> receive stop -> ok end end),
-    register(rabbit_boot, Marker),
-    try
-        StartFun()
-    catch
-        throw:{could_not_start, _App, _Reason}=Err ->
-            boot_error(Err, not_available);
-         _:Reason ->
-            boot_error(Reason, erlang:get_stacktrace())
-    after
-        unlink(Marker),
-        Marker ! stop,
-        %% give the error loggers some time to catch up
-        timer:sleep(100)
+    case catch register(rabbit_boot, Marker) of
+        true -> try
+                    case is_running() of
+                        true  -> ok;
+                        false -> StartFun()
+                    end
+                catch
+                    throw:{could_not_start, _App, _Reason}=Err ->
+                        boot_error(Err, not_available);
+                    _:Reason ->
+                        boot_error(Reason, erlang:get_stacktrace())
+                after
+                    unlink(Marker),
+                    Marker ! stop,
+                    %% give the error loggers some time to catch up
+                    timer:sleep(100)
+                end;
+        _    -> unlink(Marker),
+                Marker ! stop
     end.
 
 stop() ->
@@ -588,6 +594,7 @@ boot_delegate() ->
     rabbit_sup:start_supervisor_child(delegate_sup, [Count]).
 
 recover() ->
+    rabbit_policy:recover(),
     Qs = rabbit_amqqueue:recover(),
     ok = rabbit_binding:recover(rabbit_exchange:recover(),
                                 [QName || #amqqueue{name = QName} <- Qs]),
@@ -748,7 +755,7 @@ log_banner() ->
                     {"config file(s)" = K, []} ->
                         Format(K, "(none)");
                     {"config file(s)" = K, [V0 | Vs]} ->
-                        Format(K, V0), [Format("", V) || V <- Vs];
+                        [Format(K, V0) | [Format("", V) || V <- Vs]];
                     {K, V} ->
                         Format(K, V)
                 end || S <- Settings]),
