@@ -152,7 +152,7 @@ exists(Binding) ->
     binding_action(
       Binding, fun (_Src, _Dst, B) ->
                        rabbit_misc:const(mnesia:read({rabbit_route, B}) /= [])
-               end).
+               end, fun not_found_or_absent_errs/1).
 
 add(Binding) -> add(Binding, fun (_Src, _Dst) -> ok end).
 
@@ -177,7 +177,7 @@ add(Binding, InnerFun) ->
                   {error, _} = Err ->
                       rabbit_misc:const(Err)
               end
-      end).
+      end, fun not_found_or_absent_errs/1).
 
 add(Src, Dst, B) ->
     [SrcDurable, DstDurable] = [durable(E) || E <- [Src, Dst]],
@@ -201,13 +201,16 @@ remove(Binding, InnerFun) ->
       Binding,
       fun (Src, Dst, B) ->
               case mnesia:read(rabbit_route, B, write) of
-                  []  -> rabbit_misc:const({error, binding_not_found});
-                  [_] -> case InnerFun(Src, Dst) of
-                             ok               -> remove(Src, Dst, B);
-                             {error, _} = Err -> rabbit_misc:const(Err)
-                         end
+                  [] -> case mnesia:read(rabbit_durable_route, B, write) of
+                            [] -> rabbit_misc:const(ok);
+                            _  -> rabbit_misc:const({error, binding_not_found})
+                        end;
+                  _  -> case InnerFun(Src, Dst) of
+                            ok               -> remove(Src, Dst, B);
+                            {error, _} = Err -> rabbit_misc:const(Err)
+                        end
               end
-      end).
+      end, fun absent_errs_only/1).
 
 remove(Src, Dst, B) ->
     ok = sync_route(#route{binding = B}, durable(Src), durable(Dst),
@@ -280,9 +283,10 @@ info_all(VHostPath, Items) -> map(VHostPath, fun (B) -> info(B, Items) end).
 
 has_for_source(SrcName) ->
     Match = #route{binding = #binding{source = SrcName, _ = '_'}},
-    %% we need to check for durable routes here too in case a bunch of
-    %% routes to durable queues have been removed temporarily as a
-    %% result of a node failure
+    %% we need to check for semi-durable routes (which subsumes
+    %% durable routes) here too in case a bunch of routes to durable
+    %% queues have been removed temporarily as a result of a node
+    %% failure
     contains(rabbit_route, Match) orelse
         contains(rabbit_semi_durable_route, Match).
 
@@ -290,8 +294,9 @@ remove_for_source(SrcName) ->
     lock_route_tables(),
     Match = #route{binding = #binding{source = SrcName, _ = '_'}},
     remove_routes(
-      lists:usort(mnesia:match_object(rabbit_route, Match, write) ++
-                      mnesia:match_object(rabbit_durable_route, Match, write))).
+      lists:usort(
+        mnesia:match_object(rabbit_route, Match, write) ++
+            mnesia:match_object(rabbit_semi_durable_route, Match, write))).
 
 remove_for_destination(DstName) ->
     remove_for_destination(DstName, fun remove_routes/1).
@@ -306,13 +311,13 @@ durable(#amqqueue{durable = D}) -> D.
 
 binding_action(Binding = #binding{source      = SrcName,
                                   destination = DstName,
-                                  args        = Arguments}, Fun) ->
+                                  args        = Arguments}, Fun, ErrFun) ->
     call_with_source_and_destination(
       SrcName, DstName,
       fun (Src, Dst) ->
               SortedArgs = rabbit_misc:sort_field_table(Arguments),
               Fun(Src, Dst, Binding#binding{args = SortedArgs})
-      end).
+      end, ErrFun).
 
 delete_object(Tab, Record, LockKind) ->
     %% this 'guarded' delete prevents unnecessary writes to the mnesia
@@ -322,8 +327,6 @@ delete_object(Tab, Record, LockKind) ->
         [_] -> mnesia:delete_object(Tab, Record, LockKind)
     end.
 
-sync_route(R, Fun) -> sync_route(R, true, true, Fun).
-
 sync_route(Route, true, true, Fun) ->
     ok = Fun(rabbit_durable_route, Route, write),
     sync_route(Route, false, true, Fun);
@@ -339,13 +342,9 @@ sync_transient_route(Route, Fun) ->
     ok = Fun(rabbit_route, Route, write),
     ok = Fun(rabbit_reverse_route, reverse_route(Route), write).
 
-call_with_source_and_destination(SrcName, DstName, Fun) ->
+call_with_source_and_destination(SrcName, DstName, Fun, ErrFun) ->
     SrcTable = table_for_resource(SrcName),
     DstTable = table_for_resource(DstName),
-    ErrFun = fun (Names) ->
-                     Errs = [not_found_or_absent(Name) || Name <- Names],
-                     rabbit_misc:const({error, {resources_missing, Errs}})
-             end,
     rabbit_misc:execute_mnesia_tx_with_tail(
       fun () ->
               case {mnesia:read({SrcTable, SrcName}),
@@ -357,6 +356,18 @@ call_with_source_and_destination(SrcName, DstName, Fun) ->
               end
       end).
 
+not_found_or_absent_errs(Names) ->
+    Errs = [not_found_or_absent(Name) || Name <- Names],
+    rabbit_misc:const({error, {resources_missing, Errs}}).
+
+absent_errs_only(Names) ->
+    Errs = [E || Name <- Names,
+                 {absent, _Q} = E <- [not_found_or_absent(Name)]],
+    rabbit_misc:const(case Errs of
+                          [] -> ok;
+                          _  -> {error, {resources_missing, Errs}}
+                      end).
+
 table_for_resource(#resource{kind = exchange}) -> rabbit_exchange;
 table_for_resource(#resource{kind = queue})    -> rabbit_queue.
 
@@ -398,14 +409,17 @@ lock_route_tables() ->
 remove_routes(Routes) ->
     %% This partitioning allows us to suppress unnecessary delete
     %% operations on disk tables, which require an fsync.
-    {TransientRoutes, DurableRoutes} =
+    {RamRoutes, DiskRoutes} =
         lists:partition(fun (R) -> mnesia:match_object(
                                      rabbit_durable_route, R, write) == [] end,
                         Routes),
-    [ok = sync_transient_route(R, fun mnesia:delete_object/3) ||
-        R <- TransientRoutes],
-    [ok = sync_route(R, fun mnesia:delete_object/3) ||
-        R <- DurableRoutes],
+    %% Of course the destination might not really be durable but it's
+    %% just as easy to try to delete it from the semi-durable table
+    %% than check first
+    [ok = sync_route(R, false, true, fun mnesia:delete_object/3) ||
+        R <- RamRoutes],
+    [ok = sync_route(R, true,  true, fun mnesia:delete_object/3) ||
+        R <- DiskRoutes],
     [R#route.binding || R <- Routes].
 
 remove_transient_routes(Routes) ->
