@@ -244,10 +244,16 @@ lookup_or_die(Name) ->
         {error, not_found} -> rabbit_misc:not_found(Name)
     end.
 
+%% Not dirty_match_object since that would not be transactional when used in a
+%% tx context
 list(VHostPath) ->
-    mnesia:dirty_match_object(
-      rabbit_exchange,
-      #exchange{name = rabbit_misc:r(VHostPath, exchange), _ = '_'}).
+    mnesia:async_dirty(
+      fun () ->
+              mnesia:match_object(
+                rabbit_exchange,
+                #exchange{name = rabbit_misc:r(VHostPath, exchange), _ = '_'},
+                read)
+      end).
 
 lookup_scratch(Name, App) ->
     case lookup(Name) of
@@ -347,12 +353,11 @@ route1(Delivery, Decorators,
            lists:foldl(fun process_route/2, {WorkList, SeenXs, QNames},
                        AlternateDests ++ DecorateDests  ++ ExchangeDests)).
 
-process_alternate(#exchange{arguments = []}, _Results) -> %% optimisation
-    [];
-process_alternate(#exchange{name = XName, arguments = Args}, []) ->
-    case rabbit_misc:r_arg(XName, exchange, Args, <<"alternate-exchange">>) of
+process_alternate(X = #exchange{name = XName}, []) ->
+    case rabbit_policy:get_arg(
+           <<"alternate-exchange">>, <<"alternate-exchange">>, X) of
         undefined -> [];
-        AName     -> [AName]
+        AName     -> [rabbit_misc:r(XName, exchange, AName)]
     end;
 process_alternate(_X, _Results) ->
     [].
