@@ -197,24 +197,25 @@ start_link() -> gen_server2:start_link(?MODULE, [], []).
 
 new(Pid) ->
     %% this a 'call' to ensure that it is invoked at most once.
-    ok = gen_server:call(Pid, {new, self()}),
+    ok = gen_server:call(Pid, {new, self()}, infinity),
     #lstate{pid = Pid, prefetch_limited = false, blocked = false}.
 
 limit_prefetch(L, PrefetchCount, UnackedCount) when PrefetchCount > 0 ->
-    ok = gen_server:call(L#lstate.pid,
-                         {limit_prefetch, PrefetchCount, UnackedCount}),
+    ok = gen_server:call(
+           L#lstate.pid,
+           {limit_prefetch, PrefetchCount, UnackedCount}, infinity),
     L#lstate{prefetch_limited = true}.
 
 unlimit_prefetch(L) ->
-    ok = gen_server:call(L#lstate.pid, unlimit_prefetch),
+    ok = gen_server:call(L#lstate.pid, unlimit_prefetch, infinity),
     L#lstate{prefetch_limited = false}.
 
 block(L) ->
-    ok = gen_server:call(L#lstate.pid, block),
+    ok = gen_server:call(L#lstate.pid, block, infinity),
     L#lstate{blocked = true}.
 
 unblock(L) ->
-    ok = gen_server:call(L#lstate.pid, unblock),
+    ok = gen_server:call(L#lstate.pid, unblock, infinity),
     L#lstate{blocked = false}.
 
 is_prefetch_limited(#lstate{prefetch_limited = Limited}) -> Limited.
@@ -224,7 +225,8 @@ is_blocked(#lstate{blocked = Blocked}) -> Blocked.
 is_active(L) -> is_prefetch_limited(L) orelse is_blocked(L).
 
 get_prefetch_limit(#lstate{prefetch_limited = false}) -> 0;
-get_prefetch_limit(L) -> gen_server:call(L#lstate.pid, get_prefetch_limit).
+get_prefetch_limit(L) ->
+    gen_server:call(L#lstate.pid, get_prefetch_limit, infinity).
 
 ack(#lstate{prefetch_limited = false}, _AckCount) -> ok;
 ack(L, AckCount) -> gen_server:cast(L#lstate.pid, {ack, AckCount}).
@@ -243,9 +245,9 @@ can_send(L = #qstate{pid = Pid, state = State, credits = Credits},
     case is_consumer_blocked(L, CTag) of
         false -> case (State =/= active orelse
                        safe_call(Pid, {can_send, self(), AckRequired}, true)) of
-                     true  -> {continue, L#qstate{
-                                credits = record_send_q(CTag, Credits)}};
-                     false -> {suspend, L#qstate{state = suspended}}
+                     true  -> Credits1 = decrement_credit(CTag, Credits),
+                              {continue, L#qstate{credits = Credits1}};
+                     false -> {suspend,  L#qstate{state = suspended}}
                  end;
         true  -> {suspend, L}
     end.
@@ -269,9 +271,9 @@ is_suspended(#qstate{})                  -> false.
 
 is_consumer_blocked(#qstate{credits = Credits}, CTag) ->
     case gb_trees:lookup(CTag, Credits) of
+        none                                    -> false;
         {value, #credit{credit = C}} when C > 0 -> false;
-        {value, #credit{}}                      -> true;
-        none                                    -> false
+        {value, #credit{}}                      -> true
     end.
 
 credit(Limiter = #qstate{credits = Credits}, CTag, Credit, Drain) ->
@@ -301,7 +303,7 @@ forget_consumer(Limiter = #qstate{credits = Credits}, CTag) ->
 %% state for us (#qstate.credits), and maintain a fiction that the
 %% limiter is making the decisions...
 
-record_send_q(CTag, Credits) ->
+decrement_credit(CTag, Credits) ->
     case gb_trees:lookup(CTag, Credits) of
         {value, #credit{credit = Credit, drain = Drain}} ->
             update_credit(CTag, Credit - 1, Drain, Credits);
