@@ -25,7 +25,7 @@
 -behaviour(gen_server2).
 
 -export([start_link/0, register/2, deregister/1,
-         report_ram_duration/2, stop/0]).
+         report_ram_duration/2, stop/0, conserve_resources/3]).
 
 -export([init/1, handle_call/3, handle_cast/2, handle_info/2,
          terminate/2, code_change/3]).
@@ -36,24 +36,14 @@
                 queue_durations,      %% ets #process
                 queue_duration_sum,   %% sum of all queue_durations
                 queue_duration_count, %% number of elements in sum
-                desired_duration      %% the desired queue duration
+                desired_duration,     %% the desired queue duration
+                disk_alarm            %% disable paging, disk alarm has fired
                }).
 
 -define(SERVER, ?MODULE).
 -define(DEFAULT_UPDATE_INTERVAL, 2500).
 -define(TABLE_NAME, ?MODULE).
 
-%% Because we have a feedback loop here, we need to ensure that we
-%% have some space for when the queues don't quite respond as fast as
-%% we would like, or when there is buffering going on in other parts
-%% of the system. In short, we aim to stay some distance away from
-%% when the memory alarms will go off, which cause backpressure (of
-%% some sort) on producers. Note that all other Thresholds are
-%% relative to this scaling.
--define(MEMORY_LIMIT_SCALING, 0.4).
-
--define(LIMIT_THRESHOLD, 0.5). %% don't limit queues when mem use is < this
-
 %% If all queues are pushed to disk (duration 0), then the sum of
 %% their reported lengths will be 0. If memory then becomes available,
 %% unless we manually intervene, the sum will remain 0, and the queues
@@ -97,6 +87,11 @@ report_ram_duration(Pid, QueueDuration) ->
 stop() ->
     gen_server2:cast(?SERVER, stop).
 
+conserve_resources(Pid, disk, Conserve) ->
+    gen_server2:cast(Pid, {disk_alarm, Conserve});
+conserve_resources(_Pid, _Source, _Conserve) ->
+    ok.
+
 %%----------------------------------------------------------------------------
 %% Gen_server callbacks
 %%----------------------------------------------------------------------------
@@ -105,13 +100,14 @@ init([]) ->
     {ok, TRef} = timer:send_interval(?DEFAULT_UPDATE_INTERVAL, update),
 
     Ets = ets:new(?TABLE_NAME, [set, private, {keypos, #process.pid}]),
-
+    Alarms = rabbit_alarm:register(self(), {?MODULE, conserve_resources, []}),
     {ok, internal_update(
            #state { timer                = TRef,
                     queue_durations      = Ets,
                     queue_duration_sum   = 0.0,
                     queue_duration_count = 0,
-                    desired_duration     = infinity })}.
+                    desired_duration     = infinity,
+                    disk_alarm           = lists:member(disk, Alarms)})}.
 
 handle_call({report_ram_duration, Pid, QueueDuration}, From,
             State = #state { queue_duration_sum = Sum,
@@ -148,6 +144,12 @@ handle_call({register, Pid, MFA}, _From,
 handle_call(_Request, _From, State) ->
     {noreply, State}.
 
+handle_cast({disk_alarm, Alarm}, State = #state{disk_alarm = Alarm}) ->
+    {noreply, State};
+
+handle_cast({disk_alarm, Alarm}, State) ->
+    {noreply, internal_update(State#state{disk_alarm = Alarm})};
+
 handle_cast({deregister, Pid}, State) ->
     {noreply, internal_deregister(Pid, true, State)};
 
@@ -203,55 +205,65 @@ internal_deregister(Pid, Demonitor,
                            queue_duration_count = Count1 }
     end.
 
-internal_update(State = #state { queue_durations = Durations,
-                                 desired_duration = DesiredDurationAvg,
-                                 queue_duration_sum = Sum,
-                                 queue_duration_count = Count }) ->
-    MemoryLimit = ?MEMORY_LIMIT_SCALING * vm_memory_monitor:get_memory_limit(),
+internal_update(State = #state{queue_durations  = Durations,
+                               desired_duration = DesiredDurationAvg,
+                               disk_alarm       = DiskAlarm}) ->
+    DesiredDurationAvg1 = desired_duration_average(State),
+    ShouldInform = should_inform_predicate(DiskAlarm),
+    case ShouldInform(DesiredDurationAvg, DesiredDurationAvg1) of
+        true  -> inform_queues(ShouldInform, DesiredDurationAvg1, Durations);
+        false -> ok
+    end,
+    State#state{desired_duration = DesiredDurationAvg1}.
+
+desired_duration_average(#state{disk_alarm           = true}) ->
+    infinity;
+desired_duration_average(#state{disk_alarm           = false,
+                                queue_duration_sum   = Sum,
+                                queue_duration_count = Count}) ->
+    {ok, LimitThreshold} =
+        application:get_env(rabbit, vm_memory_high_watermark_paging_ratio),
+    MemoryLimit = vm_memory_monitor:get_memory_limit(),
     MemoryRatio = case MemoryLimit > 0.0 of
                       true  -> erlang:memory(total) / MemoryLimit;
                       false -> infinity
                   end,
-    DesiredDurationAvg1 =
-        if MemoryRatio =:= infinity ->
-                0.0;
-           MemoryRatio < ?LIMIT_THRESHOLD orelse Count == 0 ->
-                infinity;
-           MemoryRatio < ?SUM_INC_THRESHOLD ->
-                ((Sum + ?SUM_INC_AMOUNT) / Count) / MemoryRatio;
-           true ->
-                (Sum / Count) / MemoryRatio
-        end,
-    State1 = State #state { desired_duration = DesiredDurationAvg1 },
+    if MemoryRatio =:= infinity ->
+            0.0;
+       MemoryRatio < LimitThreshold orelse Count == 0 ->
+            infinity;
+       MemoryRatio < ?SUM_INC_THRESHOLD ->
+            ((Sum + ?SUM_INC_AMOUNT) / Count) / MemoryRatio;
+       true ->
+            (Sum / Count) / MemoryRatio
+    end.
 
-    %% only inform queues immediately if the desired duration has
-    %% decreased
-    case DesiredDurationAvg1 == infinity orelse
-        (DesiredDurationAvg /= infinity andalso
-         DesiredDurationAvg1 >= DesiredDurationAvg) of
-        true ->
-            ok;
-        false ->
-            true =
-                ets:foldl(
-                  fun (Proc = #process { reported = QueueDuration,
-                                         sent = PrevSendDuration,
-                                         callback = {M, F, A} }, true) ->
-                          case should_send(QueueDuration, PrevSendDuration,
-                                           DesiredDurationAvg1) of
-                              true  -> ok = erlang:apply(
-                                              M, F, A ++ [DesiredDurationAvg1]),
-                                       ets:insert(
-                                         Durations,
-                                         Proc #process {
-                                           sent = DesiredDurationAvg1});
-                              false -> true
-                          end
-                  end, true, Durations)
-    end,
-    State1.
+inform_queues(ShouldInform, DesiredDurationAvg, Durations) ->
+    true =
+        ets:foldl(
+          fun (Proc = #process{reported = QueueDuration,
+                               sent     = PrevSendDuration,
+                               callback = {M, F, A}}, true) ->
+                  case ShouldInform(PrevSendDuration, DesiredDurationAvg)
+                      andalso ShouldInform(QueueDuration, DesiredDurationAvg) of
+                      true  -> ok = erlang:apply(
+                                      M, F, A ++ [DesiredDurationAvg]),
+                               ets:insert(
+                                 Durations,
+                                 Proc#process{sent = DesiredDurationAvg});
+                      false -> true
+                  end
+          end, true, Durations).
 
-should_send(infinity, infinity,  _) -> true;
-should_send(infinity,        D, DD) -> DD < D;
-should_send(D,        infinity, DD) -> DD < D;
-should_send(D1,             D2, DD) -> DD < lists:min([D1, D2]).
+%% In normal use, we only inform queues immediately if the desired
+%% duration has decreased, we want to ensure timely paging.
+should_inform_predicate(false) -> fun greater_than/2;
+%% When the disk alarm has gone off though, we want to inform queues
+%% immediately if the desired duration has *increased* - we want to
+%% ensure timely stopping paging.
+should_inform_predicate(true) ->  fun (D1, D2) -> greater_than(D2, D1) end.
+
+greater_than(infinity, infinity) -> false;
+greater_than(infinity, _D2)      -> true;
+greater_than(_D1,      infinity) -> false;
+greater_than(D1,       D2)       -> D1 > D2.
