@@ -17,9 +17,10 @@
 -module(rabbit_mirror_queue_misc).
 -behaviour(rabbit_policy_validator).
 
--export([remove_from_queue/3, on_node_up/0, add_mirrors/2, add_mirror/2,
+-export([remove_from_queue/3, on_node_up/0, add_mirrors/3,
          report_deaths/4, store_updated_slaves/1, suggested_queue_nodes/1,
-         is_mirrored/1, update_mirrors/2, validate_policy/1]).
+         is_mirrored/1, update_mirrors/2, validate_policy/1,
+         maybe_auto_sync/1]).
 
 %% for testing only
 -export([module/1]).
@@ -45,10 +46,8 @@
         (rabbit_amqqueue:name(), pid(), [pid()])
         -> {'ok', pid(), [pid()]} | {'error', 'not_found'}).
 -spec(on_node_up/0 :: () -> 'ok').
--spec(add_mirrors/2 :: (rabbit_amqqueue:name(), [node()]) -> 'ok').
--spec(add_mirror/2 ::
-        (rabbit_amqqueue:name(), node()) ->
-                           {'ok', atom()} | rabbit_types:error(any())).
+-spec(add_mirrors/3 :: (rabbit_amqqueue:name(), [node()], 'sync' | 'async')
+                       -> 'ok').
 -spec(store_updated_slaves/1 :: (rabbit_types:amqqueue()) ->
                                      rabbit_types:amqqueue()).
 -spec(suggested_queue_nodes/1 :: (rabbit_types:amqqueue()) ->
@@ -56,6 +55,7 @@
 -spec(is_mirrored/1 :: (rabbit_types:amqqueue()) -> boolean()).
 -spec(update_mirrors/2 ::
         (rabbit_types:amqqueue(), rabbit_types:amqqueue()) -> 'ok').
+-spec(maybe_auto_sync/1 :: (rabbit_types:amqqueue()) -> 'ok').
 
 -endif.
 
@@ -69,7 +69,7 @@
 %% slave (now master) receives messages it's not ready for (for
 %% example, new consumers).
 %% Returns {ok, NewMPid, DeadPids}
-remove_from_queue(QueueName, Self, DeadGMPids) ->
+remove_from_queue(QueueName, Self, LiveGMPids) ->
     rabbit_misc:execute_mnesia_transaction(
       fun () ->
               %% Someone else could have deleted the queue before we
@@ -79,9 +79,9 @@ remove_from_queue(QueueName, Self, DeadGMPids) ->
                   [Q = #amqqueue { pid        = QPid,
                                    slave_pids = SPids,
                                    gm_pids    = GMPids }] ->
-                      {Dead, GMPids1} = lists:partition(
+                      {GMPids1, Dead} = lists:partition(
                                           fun ({GM, _}) ->
-                                                  lists:member(GM, DeadGMPids)
+                                                  lists:member(GM, LiveGMPids)
                                           end, GMPids),
                       DeadPids = [Pid || {_GM, Pid} <- Dead],
                       Alive = [QPid | SPids] -- DeadPids,
@@ -94,10 +94,15 @@ remove_from_queue(QueueName, Self, DeadGMPids) ->
                               %% Either master hasn't changed, so
                               %% we're ok to update mnesia; or we have
                               %% become the master.
-                              store_updated_slaves(
-                                Q #amqqueue { pid        = QPid1,
+                              Q1 = Q#amqqueue{pid        = QPid1,
                                               slave_pids = SPids1,
-                                              gm_pids    = GMPids1 }),
+                                              gm_pids    = GMPids1},
+                              store_updated_slaves(Q1),
+                              %% If we add and remove nodes at the same time we
+                              %% might tell the old master we need to sync and
+                              %% then shut it down. So let's check if the new
+                              %% master needs to sync.
+                              maybe_auto_sync(Q1),
                               {ok, QPid1, [QPid | SPids] -- Alive};
                           _ ->
                               %% Master has changed, and we're not it,
@@ -135,7 +140,7 @@ on_node_up() ->
                             end
                     end, [], rabbit_queue)
           end),
-    [add_mirror(QName, node()) || QName <- QNames],
+    [add_mirror(QName, node(), async) || QName <- QNames],
     ok.
 
 drop_mirrors(QName, Nodes) ->
@@ -160,52 +165,40 @@ drop_mirror(QName, MirrorNode) ->
               end
       end).
 
-add_mirrors(QName, Nodes) ->
-    [add_mirror(QName, Node)  || Node <- Nodes],
+add_mirrors(QName, Nodes, SyncMode) ->
+    [add_mirror(QName, Node, SyncMode)  || Node <- Nodes],
     ok.
 
-add_mirror(QName, MirrorNode) ->
+add_mirror(QName, MirrorNode, SyncMode) ->
     rabbit_amqqueue:with(
       QName,
       fun (#amqqueue { name = Name, pid = QPid, slave_pids = SPids } = Q) ->
               case [Pid || Pid <- [QPid | SPids], node(Pid) =:= MirrorNode] of
                   [] ->
-                      start_child(Name, MirrorNode, Q);
+                      start_child(Name, MirrorNode, Q, SyncMode);
                   [SPid] ->
                       case rabbit_misc:is_process_alive(SPid) of
                           true  -> {ok, already_mirrored};
-                          false -> start_child(Name, MirrorNode, Q)
+                          false -> start_child(Name, MirrorNode, Q, SyncMode)
                       end
               end
       end).
 
-start_child(Name, MirrorNode, Q) ->
+start_child(Name, MirrorNode, Q, SyncMode) ->
     case rabbit_misc:with_exit_handler(
-           rabbit_misc:const({ok, down}),
+           rabbit_misc:const(down),
            fun () ->
                    rabbit_mirror_queue_slave_sup:start_child(MirrorNode, [Q])
            end) of
-        {ok, SPid} when is_pid(SPid)  ->
-            maybe_auto_sync(Q),
-            rabbit_log:info("Adding mirror of ~s on node ~p: ~p~n",
-                            [rabbit_misc:rs(Name), MirrorNode, SPid]),
-            {ok, started};
-        {error, {{stale_master_pid, StalePid}, _}} ->
-            rabbit_log:warning("Detected stale HA master while adding "
-                               "mirror of ~s on node ~p: ~p~n",
-                               [rabbit_misc:rs(Name), MirrorNode, StalePid]),
-            {ok, stale_master};
-        {error, {{duplicate_live_master, _}=Err, _}} ->
-            Err;
-        Other ->
-            Other
+        {ok, SPid} -> rabbit_log:info("Adding mirror of ~s on node ~p: ~p~n",
+                                      [rabbit_misc:rs(Name), MirrorNode, SPid]),
+                      rabbit_mirror_queue_slave:go(SPid, SyncMode);
+        _          -> ok
     end.
 
 report_deaths(_MirrorPid, _IsMaster, _QueueName, []) ->
     ok;
 report_deaths(MirrorPid, IsMaster, QueueName, DeadPids) ->
-    rabbit_event:notify(queue_mirror_deaths, [{name, QueueName},
-                                              {pids, DeadPids}]),
     rabbit_log:info("Mirrored-queue (~s): ~s ~s saw deaths of mirrors ~s~n",
                     [rabbit_misc:rs(QueueName),
                      case IsMaster of
@@ -223,7 +216,7 @@ store_updated_slaves(Q = #amqqueue{slave_pids      = SPids,
     Q1 = Q#amqqueue{sync_slave_pids = SSPids1},
     ok = rabbit_amqqueue:store_queue(Q1),
     %% Wake it up so that we emit a stats event
-    rabbit_amqqueue:wake_up(Q1),
+    rabbit_amqqueue:notify_policy_changed(Q1),
     Q1.
 
 %%----------------------------------------------------------------------------
@@ -239,28 +232,32 @@ suggested_queue_nodes(Q) ->
 %% This variant exists so we can pull a call to
 %% rabbit_mnesia:cluster_nodes(running) out of a loop or
 %% transaction or both.
-suggested_queue_nodes(Q, All) ->
+suggested_queue_nodes(Q = #amqqueue{exclusive_owner = Owner}, All) ->
     {MNode0, SNodes, SSNodes} = actual_queue_nodes(Q),
     MNode = case MNode0 of
                 none -> node();
                 _    -> MNode0
             end,
-    Params = policy(<<"ha-params">>, Q),
-    case module(Q) of
-        {ok, M} -> M:suggested_queue_nodes(Params, MNode, SNodes, SSNodes, All);
-        _       -> {MNode, []}
+    case Owner of
+        none -> Params = policy(<<"ha-params">>, Q),
+                case module(Q) of
+                    {ok, M} -> M:suggested_queue_nodes(
+                                 Params, MNode, SNodes, SSNodes, All);
+                    _       -> {MNode, []}
+                end;
+        _    -> {MNode, []}
     end.
 
 policy(Policy, Q) ->
     case rabbit_policy:get(Policy, Q) of
-        {ok, P} -> P;
-        _       -> none
+        undefined -> none;
+        P         -> P
     end.
 
 module(#amqqueue{} = Q) ->
     case rabbit_policy:get(<<"ha-mode">>, Q) of
-        {ok, Mode} -> module(Mode);
-        _          -> not_mirrored
+        undefined -> not_mirrored;
+        Mode      -> module(Mode)
     end;
 
 module(Mode) when is_binary(Mode) ->
@@ -310,8 +307,10 @@ update_mirrors0(OldQ = #amqqueue{name = QName},
     {NewMNode, NewSNodes}    = suggested_queue_nodes(NewQ),
     OldNodes = [OldMNode | OldSNodes],
     NewNodes = [NewMNode | NewSNodes],
-    add_mirrors (QName, NewNodes -- OldNodes),
+    add_mirrors (QName, NewNodes -- OldNodes, async),
     drop_mirrors(QName, OldNodes -- NewNodes),
+    %% This is for the case where no extra nodes were added but we changed to
+    %% a policy requiring auto-sync.
     maybe_auto_sync(NewQ),
     ok.
 
