@@ -396,10 +396,11 @@ cluster_status(WhichNodes) ->
 
 node_info() ->
     {erlang:system_info(otp_release), rabbit_misc:version(),
-     delegate_beam_hash(), cluster_status_from_mnesia()}.
+     cluster_status_from_mnesia()}.
 
 node_type() ->
-    DiscNodes = cluster_nodes(disc),
+    {_AllNodes, DiscNodes, _RunningNodes} =
+        rabbit_node_monitor:read_cluster_status(),
     case DiscNodes =:= [] orelse me_in_nodes(DiscNodes) of
         true  -> disc;
         false -> ram
@@ -433,10 +434,8 @@ init_db(ClusterNodes, NodeType, CheckOtherNodes) ->
             %% First disc node up
             maybe_force_load(),
             ok;
-        {[AnotherNode | _], _, _} ->
+        {[_ | _], _, _} ->
             %% Subsequent node in cluster, catch up
-            ensure_version_ok(
-              rpc:call(AnotherNode, rabbit_version, recorded, [])),
             maybe_force_load(),
             ok = rabbit_table:wait_for_replicated(),
             ok = rabbit_table:create_local_copy(NodeType)
@@ -459,10 +458,11 @@ init_db_and_upgrade(ClusterNodes, NodeType, CheckOtherNodes) ->
     %% about the cluster
     case NodeType of
         ram  -> start_mnesia(),
-                change_extra_db_nodes(ClusterNodes, false),
-                rabbit_table:wait_for_replicated();
+                change_extra_db_nodes(ClusterNodes, false);
         disc -> ok
     end,
+    %% ...and all nodes will need to wait for tables
+    rabbit_table:wait_for_replicated(),
     ok.
 
 init_db_with_mnesia(ClusterNodes, NodeType,
@@ -518,7 +518,7 @@ copy_db(Destination) ->
     rabbit_file:recursive_copy(dir(), Destination).
 
 force_load_filename() ->
-    filename:join(rabbit_mnesia:dir(), "force_load").
+    filename:join(dir(), "force_load").
 
 force_load_next_boot() ->
     rabbit_file:write_file(force_load_filename(), <<"">>).
@@ -569,16 +569,16 @@ check_cluster_consistency(Node) ->
     case rpc:call(Node, rabbit_mnesia, node_info, []) of
         {badrpc, _Reason} ->
             {error, not_found};
-        {_OTP, _Rabbit, _Hash, {error, _}} ->
+        {_OTP, _Rabbit, {error, _}} ->
             {error, not_found};
-        {_OTP, Rabbit, _Status} ->
-             %% pre-2013/04 format implies version mismatch
-            version_error("Rabbit", rabbit_misc:version(), Rabbit);
-        {OTP, Rabbit, Hash, {ok, Status}} ->
-            case check_consistency(OTP, Rabbit, Hash, Node, Status) of
+        {OTP, Rabbit, {ok, Status}} ->
+            case check_consistency(OTP, Rabbit, Node, Status) of
                 {error, _} = E -> E;
                 {ok, Res}      -> {ok, Res}
-            end
+            end;
+        {_OTP, Rabbit, _Hash, _Status} ->
+            %% delegate hash checking implies version mismatch
+            version_error("Rabbit", rabbit_misc:version(), Rabbit)
     end.
 
 %%--------------------------------------------------------------------
@@ -637,15 +637,6 @@ schema_ok_or_move() ->
             ok = create_schema()
     end.
 
-ensure_version_ok({ok, DiscVersion}) ->
-    DesiredVersion = rabbit_version:desired(),
-    case rabbit_version:matches(DesiredVersion, DiscVersion) of
-        true  -> ok;
-        false -> throw({error, {version_mismatch, DesiredVersion, DiscVersion}})
-    end;
-ensure_version_ok({error, _}) ->
-    ok = rabbit_version:record_desired().
-
 %% We only care about disc nodes since ram nodes are supposed to catch
 %% up only
 create_schema() ->
@@ -742,17 +733,15 @@ change_extra_db_nodes(ClusterNodes0, CheckOtherNodes) ->
             Nodes
     end.
 
-check_consistency(OTP, Rabbit, Hash) ->
+check_consistency(OTP, Rabbit) ->
     rabbit_misc:sequence_error(
       [check_otp_consistency(OTP),
-       check_rabbit_consistency(Rabbit),
-       check_beam_compatibility(Hash)]).
+       check_rabbit_consistency(Rabbit)]).
 
-check_consistency(OTP, Rabbit, Hash, Node, Status) ->
+check_consistency(OTP, Rabbit, Node, Status) ->
     rabbit_misc:sequence_error(
       [check_otp_consistency(OTP),
        check_rabbit_consistency(Rabbit),
-       check_beam_compatibility(Hash),
        check_nodes_consistency(Node, Status)]).
 
 check_nodes_consistency(Node, RemoteStatus = {RemoteAllNodes, _, _}) ->
@@ -783,31 +772,11 @@ version_error(Name, This, Remote) ->
 check_otp_consistency(Remote) ->
     check_version_consistency(erlang:system_info(otp_release), Remote, "OTP").
 
-%% Unlike the rest of 3.0.x, 3.0.0 is not compatible. This can be
-%% removed after 3.1.0 is released.
-check_rabbit_consistency("3.0.0") ->
-    version_error("Rabbit", rabbit_misc:version(), "3.0.0");
-
 check_rabbit_consistency(Remote) ->
     check_version_consistency(
       rabbit_misc:version(), Remote, "Rabbit",
       fun rabbit_misc:version_minor_equivalent/2).
 
-check_beam_compatibility(RemoteHash) ->
-    case RemoteHash == delegate_beam_hash() of
-        true  -> ok;
-        false -> {error, {incompatible_bytecode,
-                          "Incompatible Erlang bytecode found on nodes"}}
-    end.
-
-%% The delegate module sends functions across the cluster; if it is
-%% out of sync (say due to mixed compilers), we will get badfun
-%% exceptions when trying to do so. Let's detect that at startup.
-delegate_beam_hash() ->
-    {delegate, Obj, _} = code:get_object_code(delegate),
-    {ok, {delegate, Hash}} = beam_lib:md5(Obj),
-    Hash.
-
 %% This is fairly tricky.  We want to know if the node is in the state
 %% that a `reset' would leave it in.  We cannot simply check if the
 %% mnesia tables aren't there because restarted RAM nodes won't have
@@ -833,12 +802,13 @@ find_good_node([]) ->
     none;
 find_good_node([Node | Nodes]) ->
     case rpc:call(Node, rabbit_mnesia, node_info, []) of
-        {badrpc, _Reason}      -> find_good_node(Nodes);
-        {_OTP, _Rabbit, _}     -> find_good_node(Nodes);
-        {OTP, Rabbit, Hash, _} -> case check_consistency(OTP, Rabbit, Hash) of
-                                       {error, _} -> find_good_node(Nodes);
-                                       ok         -> {ok, Node}
-                                  end
+        {badrpc, _Reason}         -> find_good_node(Nodes);
+        %% old delegate hash check
+        {_OTP, _Rabbit, _Hash, _} -> find_good_node(Nodes);
+        {OTP, Rabbit, _}          -> case check_consistency(OTP, Rabbit) of
+                                         {error, _} -> find_good_node(Nodes);
+                                         ok         -> {ok, Node}
+                                     end
     end.
 
 is_only_clustered_disc_node() ->
